# This software was developed by employees of the US Naval Research Laboratory (NRL), an
# agency of the Federal Government. Pursuant to title 17 section 105 of the United States
# Code, works of NRL employees are not subject to copyright protection, and this software
# is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
# responsibility whatsoever for its use by other parties, and makes no guarantees,
# expressed or implied, about its quality, reliability, or any other characteristic. We
# would appreciate acknowledgment if the software is used. To the extent that NRL may hold
# copyright in countries other than the United States, you are hereby granted the
# non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
# works and distribute this software, in any medium, or authorize others to do so on your
# behalf, on a royalty-free basis throughout the world. You may improve, modify, and
# create derivative works of the software or any portion of the software, and you may copy
# and distribute such modifications or works. Modified works should carry a notice stating
# that you changed the software and should note the date and nature of any such change.
# Please explicitly acknowledge the US Naval Research Laboratory as the original source.
# This software can be redistributed and/or modified freely provided that any derivative
# works bear some notice that they are derived from it, and any modified versions bear
# some notice that they have been modified.
#
#
# Author: David Rowenhorst;
# The US Naval Research Laboratory Date: 22 May 2024
#
# For further information see:
# David J. Rowenhorst, Patrick G. Callahan, Håkon W. Ånes. Fast Radon transforms for
# high-precision EBSD orientation determination using PyEBSDIndex.
# Journal of Applied Crystallography, 57(1):3–19, 2024.
# DOI: 10.1107/S1600576723010221
#
#

"""Class that reads in EBSD patterns, and finds the bands within those patterns using a Radon Transform."""

#from os import environ
import os
from pathlib import PurePath, Path
import platform
# import tempfile
from timeit import default_timer as timer

import matplotlib.pyplot as plt
import numba
import numpy as np

import scipy.ndimage as scipyndim #import gaussian_filter, dilation ...
#from scipy.ndimage #import grey_dilation as scipy_grey_dilation
#from scipy.ndimage #import median_filter
import scipy.optimize as scipyopt

from pyebsdindex import radon_fast

tempdir = PurePath(Path.home())
#tempdir = PurePath("/tmp" if platform.system() == "Darwin" else tempfile.gettempdir())
#tempdir = tempdir.joinpath('numbacache')
tempdir = tempdir.joinpath('.pyebsdindex').joinpath('numbacache')
Path(tempdir).mkdir(parents=True, exist_ok=True)
os.environ["NUMBA_CACHE_DIR"] = str(tempdir)+str(os.sep)

RADEG = 180.0/np.pi


class BandDetect():
  def __init__(
    self,
    patterns=None,
    patDim=None,
    nTheta=180,
    nRho=90,
    tSigma=None,
    rSigma=None,
    rhoMaskFrac=0.1,
    nBands=9,
    **kwargs
):
    self.patDim = None
    self.nTheta = nTheta
    self.nRho = nRho
    self.dTheta = None
    self.dRho = None
    self.rhoMax = None
    self.radonPlan = None
    self.rdnNorm = None
    self.tSigma = tSigma
    self.rSigma = rSigma
    self.kernel = None
    self.peakPad = np.array([11, 11])
    self.padding = np.array([11, 11])
    self.rhoMaskFrac = rhoMaskFrac

    self.rhomask1thresh = None
    self.rhomask1 = None

    self.nBands = nBands
    self.EDAXIQ = False
    self.backgroundsub = None
    self.patternmask = None
    self.useCPU = True

    self.dataType = np.dtype([('id', np.int32), ('max', np.float32), \
                    ('maxloc', np.float32, (2)), ('avemax', np.float32), ('aveloc', np.float32, (2)),\
                    ('pqmax', np.float32), ('width', np.float32), ('theta', np.float32), ('rho', np.float32),
                    ('valid', np.int8),('band_match_index', np.int32, (2))])


    if (patterns is None) and (patDim is None):
      pass
    else:
      if (patterns is not None):
        self.patDim = np.asarray(patterns.shape[-2:])
      else:
        self.patDim = np.asarray(patDim)
      patternmask = None
      if 'patternmask' in kwargs :
        patternmask = kwargs.get('patternmask')

      patternmaskindex = None
      if 'patternmaskindex' in kwargs:
        patternmaskindex = kwargs.get('patternmaskindex')
      #print(patternmask)
      self.band_detect_setup(patterns, self.patDim,self.nTheta,self.nRho,\
                             self.tSigma, self.rSigma,self.rhoMaskFrac,self.nBands,
                             patternmask = patternmask,patternmaskindex = patternmaskindex,
                             **kwargs)

  def band_detect_setup(self, patterns=None,patDim=None,nTheta=None,nRho=None,\
                      tSigma=None, rSigma=None,rhoMaskFrac=None,nBands=None,
                      patternmask=None, patternmaskindex=None, **kwargs):
    p_dim = None
    recalc_radon = False
    recalc_masks = False
    if (patterns is None) and (patDim is not None):
      p_dim = np.asarray(patDim, dtype=np.int64)
    if patterns is not None:
      p_dim = np.shape(patterns)[-2:]  # this will catch if someone sends in a [1 x N x M] image
    if p_dim is not None:
      if self.patDim is None:
        recalc_radon = True
        self.patDim = p_dim

      elif np.sum(np.abs(self.patDim[-2:]-p_dim[-2:]), dtype=np.int64) != 0:
        recalc_radon = True
        self.patDim = p_dim

    if nTheta is not None:
      self.nTheta = nTheta
      recalc_radon = True
      recalc_masks = True

    if self.nTheta is not None:
      self.dTheta = 180.0/self.nTheta


    if nRho is not None:
      self.nRho = nRho
      recalc_radon = True
      recalc_masks = True

    if self.dRho is None:
      recalc_radon = True

    #recalc_radon = True
    if recalc_radon == True:
      if (self.rhoMaskFrac < 1) and (self.rhoMaskFrac > 0):
        self.rhoMax = 0.5 * np.float32(self.patDim.min())
      else:
        self.rhoMax = 0.5 * np.float32(np.sqrt(np.float32(self.patDim[-2])**2  + np.float32(self.patDim[-1])**2))

      self.rhomask1thresh = np.float32(self.patDim.min())*0.1

      self.dRho = self.rhoMax/np.float32(self.nRho)
      self.radonPlan = radon_fast.Radon(imageDim=self.patDim,
                                        nTheta=self.nTheta, nRho=self.nRho,
                                        rhoMax=self.rhoMax,
                                        mask=patternmask, maskindex=patternmaskindex)

      if patternmask is not None:
        back = np.array(patternmask > 0).astype(np.float32)
      else:
        back = np.ones(self.patDim[-2:], dtype=np.float32)



      rdnmask = self.radonPlan.radon_faster(back,fixArtifacts=False, normalization=False)
      #plt.imshow(rdnmask >= self.rhomask1thresh)
      self.rhomask1 = (rdnmask[:,:,0] >= self.rhomask1thresh).astype(np.ubyte)

      rdnmask = rdnmask > 0
      rdnmask = rdnmask.squeeze()

      if self.rhoMaskFrac >= 1:
        s = np.ones(( 3, 1))
        rdnmask = scipyndim.binary_erosion(rdnmask, structure=s, iterations = int(self.rhoMaskFrac) )
      else:
        cmask = self._circmask(back)
        rdncmask = rdnmask.copy()
        if self.rhoMaskFrac > 0:
          mskinx = int(self.nRho*self.rhoMaskFrac)
          if mskinx == 0:
            mskinx = 1
          rdncmask[0:mskinx, :] = 0
          rdncmask[-mskinx:, :] = 0

          #thresh = 0.5 * (np.min(back.shape[-2:]) * (1.0 - self.rhoMaskFrac))
          #cmask = (cmask < thresh).astype(np.float32)
        elif self.rhoMaskFrac < 0:
          mskinx = int(-1*self.nRho * self.rhoMaskFrac)
          #rdncmask[0:mskinx, :] = 0
          #rdncmask[-mskinx:, :] = 0

          cmask = (cmask < self.rhoMax * (1.0 + self.rhoMaskFrac)).astype(np.float32)
          rdncmask = (self.radonPlan.radon_faster(cmask, fixArtifacts=False) > 0).squeeze()
          #plt.imshow(cmask); print(thresh)
        else:
          pass
          #cmask[:,...] = (cmask > -1).astype(np.float32)


        #rdncmask = (self.radonPlan.radon_faster(cmask, fixArtifacts=False) > 0).squeeze()

        rdnmask *= (rdncmask > 0).astype(bool)



      self.rdnmask = rdnmask > 0
      #back = (back > 0).astype(np.float32) / (back + 1.0e-12)
      #self.rdnNorm = back


    if tSigma is not None:
      self.tSigma = tSigma
      recalc_masks = True
    if rSigma is not None:
      self.rSigma = rSigma
      recalc_masks = True

    if rhoMaskFrac is not None:
      self.rhoMaskFrac = rhoMaskFrac
      recalc_masks = True
    if (self.rhoMaskFrac is not None):
      recalc_masks = True

    if (self.tSigma is None) and (self.dTheta is not None):
      self.tSigma = 1.0/self.dTheta
      recalc_masks = True

    if (self.rSigma is None) and (self.dRho is not None):
      self.rSigma = 0.5/np.float32(self.dRho)
      recalc_masks = True

    if recalc_masks == True:
      ksz = np.array([np.max([np.int64(4*self.rSigma), 5]), np.max([np.int64(4*self.tSigma), 5])])
      ksz = ksz + ((ksz % 2) == 0)
      kernel = np.zeros(ksz, dtype=np.float32)
      kernel[(ksz[0]/2).astype(int),(ksz[1]/2).astype(int) ] = 1
      kernel = -1.0*scipyndim.gaussian_filter(kernel, [self.rSigma, self.tSigma], order=[2,0])
      self.kernel = kernel.reshape((1,ksz[0], ksz[1]))
      #self.peakPad = np.array(np.around([ 4*ksz[0], 20.0/self.dTheta]), dtype=np.int64)
      self.peakPad = np.array(np.around([2 * ksz[0], 2 * ksz[1]]), dtype=np.int64)
      self.peakPad += 1 - np.mod(self.peakPad, 2)  # make sure we have it as odd.

    self.padding = np.array([np.max( [self.peakPad[0], self.padding[0]] ), np.max([self.peakPad[1], self.padding[1]])])

    if nBands is not None:
      self.nBands = nBands

  def collect_background(self, fileobj = None, patsIn = None, nsample = None, method = 'randomStride', sigma=None):

    back = None # default value
    # we got an array of patterns

    if patsIn is not None:
      ndim = patsIn.ndim
      if ndim == 2:
        patsIn = np.expand_dims(patsIn,axis=0)
      else:
        patsIn = patsIn
      npats = patsIn.shape[0]
      if nsample is None:
        nsample = npats
      #pshape = patsIn.shape
      if npats <= nsample:
        back = np.mean(patsIn, axis = 0)
        back = np.expand_dims(back,axis=0)
      else:
        if method.upper() == 'RANDOMSTRIDE':
          stride = np.random.choice(npats, size = nsample, replace = False )
          stride = np.sort(stride)
          back = np.mean(patsIn[stride,:,:],axis=0)
        elif method.upper() == 'EVENSTRIDE':
          stride = np.arange(0, npats, int(npats/nsample)) # not great, but maybe good enough.
          back = np.mean(patsIn[stride, :, :], axis=0)

    if (back is None) and (fileobj is not None):
      if fileobj.version is None:
        fileobj.read_header()
      npats = fileobj.nPatterns
      if nsample is None:
        nsample = npats
      if npats <= nsample:
        nsample = npats

      if method.upper() == 'RANDOMSTRIDE':
        stride = np.random.choice(npats, size = nsample, replace = False )
        stride = np.sort(stride)
      elif method.upper() == 'EVENSTRIDE':
        step = int(npats / nsample) # not great, but maybe good enough.
        stride = np.arange(0,npats, step, dypte = np.uint64)
      pat1 = fileobj.read_data(convertToFloat=True,patStartCount=[stride[0], 1],returnArrayOnly=True)[0]
      for i in stride[1:]:
        pat1 += fileobj.read_data(convertToFloat=True,patStartCount=[i, 1],returnArrayOnly=True)[0]
      back = pat1 / float(len(stride))
      #pshape = pat1.shape
    # a bit of image processing.
    if back is not None:
      back = np.squeeze(back)
      back = self.backsub_fit(back)
    self.backgroundsub = back

  def backsub_fit(self, back, mask = None):
    # This function will fit a 2D gaussian on top of a plane to the averaged set of patterns (data) that is provided.
    # It will automatically use whatever mask is defined for valid data.
    # If the gaussian fit fails to converge, it will fall back to just using the mean set of patterns for the background
    # with a warning.
    def gaussian_surf(x, y, a, x0, y0, sigx, sigy, c, d, e):
    # equation for 2D gaussian on top of a plane.
      return a * np.exp(- ((x - x0) ** 2) / (2.0 * sigx ** 2) - ((y - y0) ** 2) / (2.0 * sigy ** 2)) + c + d * x + e * y

    def fit_gauss(M, *args):
    # helper function
      x, y = M
      #arr = np.zeros(x.shape)
      return gaussian_surf(x, y, *args)

    #back = np.mean(data, axis=0) # start with the mean of all the data
    # now fit a 2D gaussian sitting on a plane.  See fuction def above.
    nx = back.shape[-1]
    ny = back.shape[-2]
    #plt.imshow(back)
    x = np.arange(nx, dtype=float)
    x = (np.broadcast_to(x.reshape(1,nx), (ny, nx))).ravel()
    y = np.arange(ny, dtype=float)
    y = (np.broadcast_to(y, (nx, ny)).T).ravel()
    if mask is None:
      # make a circular mask - even if not EDAX, this should work OK.
      cx = (np.arange(nx) - nx*0.5)**2
      cy = (np.arange(ny) - ny*0.5)**2
      cmask = np.sqrt(np.broadcast_to(cx.reshape(1,nx), (ny, nx)) + np.broadcast_to(cy, (nx, ny)).T) < (ny*0.49)
    else:
      cmask = mask
    # need to grab only the values that are in the mask.
    wh = np.nonzero(cmask.ravel())[0]
    xwh = x[wh]
    ywh = y[wh]
    xywh = np.vstack((xwh, ywh))
    zwh = (scipyndim.median_filter(np.squeeze(back),3).ravel())[wh]
    whmx = np.unravel_index(back.argmax(), back.shape)
    minz = zwh.min()
    # initialize a guess for the parameters.
    # [gauss amplitude, max loc x, max loc y, sigx, sigy, const offset, slope x, slope y]
    p0 = [(zwh.max() - zwh.min())*0.1, whmx[1], whmx[0], nx/2.355, ny/2.355, minz, 0, 0]
    try:
      popt, pcov = scipyopt.curve_fit(fit_gauss, xywh, zwh, p0)
      backfit = (gaussian_surf(x, y, *popt)).reshape(ny, nx)
      #print(p0, popt)
    except RuntimeError:
      print('Warning: no convergence on background gaussian fit ... using mean of the patterns.')
      print('This may not be ideal for scans with few grains across the width of the scan.')
      backfit = back
    backfit -= np.mean(backfit)
    #f, axarr = plt.subplots(1, 3)
    #f.set_size_inches(10, 4)
    #axarr[0].imshow(data[0,:,:].squeeze(), cmap='gray')
    #axarr[1].imshow(data[0,:,:].squeeze() - backfit, cmap='gray')
    #axarr[2].imshow(backfit, cmap='gray')


    return backfit


  def find_bands(self, patternsIn, verbose=0, chunksize=-1,  **kwargs):
    tic0 = timer()
    tic = timer()
    ndim = patternsIn.ndim
    if ndim == 2:
      patterns = np.expand_dims(patternsIn, axis=0)
    else:
      patterns = patternsIn

    shape = patterns.shape
    nPats = shape[0]

    bandData = np.zeros((nPats,self.nBands),dtype=self.dataType)
    bandData['band_match_index'] = -100
    if chunksize < 0:
      nchunks = 1
      chunksize = nPats
      chunk_start_end = [[0,nPats]]
    else:
      nchunks = (np.ceil(nPats / chunksize)).astype(np.compat.long)
      chunk_start_end = [[i * chunksize, (i + 1) * chunksize] for i in range(nchunks)]
      chunk_start_end[-1][1] = nPats

    # these are timers used to gauge performance
    rdntime = 0.0
    convtime = 0.0
    lmaxtime = 0.0
    blabeltime = 0.0

    for chnk in chunk_start_end:
      tic1 = timer()
      rdnNorm = self.radonPlan.radon_faster(patterns[chnk[0]:chnk[1],:,:], self.padding, fixArtifacts=False, background=self.backgroundsub)
      rdntime += timer() - tic1
      tic1 = timer()
      rdnConv = self.rdn_conv(rdnNorm)
      convtime += timer()-tic1
      tic1 = timer()
      lMaxRdn= self.rdn_local_max(rdnConv)
      lmaxtime +=  timer()-tic1
      tic1 = timer()
      bandDataChunk= self.band_label(chnk[1]-chnk[0], rdnConv, rdnNorm, lMaxRdn)
      bandData[chnk[0]:chnk[1]] = bandDataChunk
      if (verbose > 1) and (chnk[1] == nPats): # need to pull the radonconv off the gpu
        rdnConv = rdnConv[:,:,0:chnk[1]-chnk[0] ]

      blabeltime += timer() - tic1

    tottime = timer() - tic0

    if verbose > 0:
      print('Radon Time:',rdntime)
      print('Convolution Time:', convtime)
      print('Peak ID Time:', lmaxtime)
      print('Band Label Time:', blabeltime)
      print('Total Band Find Time:',tottime)
    if verbose > 1:
      self._display_radon_pattern(rdnConv, bandData, patterns)
      # if len(rdnConv.shape) == 3:
      #   im2show = rdnConv[self.padding[0]:-self.padding[0],self.padding[1]:-self.padding[1], -1]
      # else:
      #   im2show = rdnConv[self.padding[0]:-self.padding[0],self.padding[1]:-self.padding[1]]
      #
      # rhoMaskTrim = np.int32(im2show.shape[0] * self.rhoMaskFrac)
      # mean = np.mean(im2show[rhoMaskTrim:-rhoMaskTrim, 1:-2])
      # stdv = np.std(im2show[rhoMaskTrim:-rhoMaskTrim, 1:-2])
      #
      # im2show -= mean
      # im2show /= stdv
      # im2show = im2show.clip(-4, None)
      # im2show += 6
      # im2show[0:rhoMaskTrim,:] = 0
      # im2show[-rhoMaskTrim:,:] = 0
      # im2show = np.fliplr(im2show)
      #
      # plt.figure()
      # plt.imshow(im2show, cmap='gray', extent=[self.radonPlan.theta.min(), self.radonPlan.theta.max(),
      #                                          self.radonPlan.rho.min(), self.radonPlan.rho.max()],
      #            interpolation='none', zorder=1, aspect='auto')
      # width = bandData['width'][-1, :]
      # width /= width.min()
      # width *= 2
      # xplt = np.squeeze(
      #   180.0 - np.interp(bandData['aveloc'][-1, :, 1]+0.5, np.arange(self.radonPlan.nTheta), self.radonPlan.theta))
      # yplt = np.squeeze(
      #   -1.0 * np.interp(bandData['aveloc'][-1, :, 0]-0.5, np.arange(self.radonPlan.nRho), self.radonPlan.rho))
      #
      # plt.scatter(y=yplt, x=xplt, c='r', s=width, zorder=2)
      #
      # for pt in range(self.nBands):
      #   plt.annotate(str(pt + 1), np.squeeze([xplt[pt]+4, yplt[pt]]), color='yellow')
      # plt.xlim(0,180)
      # plt.ylim(-self.rhoMax, self.rhoMax)


    return bandData

  def radonPad(self,radon,rPad=0,tPad = 0, mirrorTheta = True):
    # function for padding the radon transform
    # theta padding (tPad) will use the symmetry of the radon and will vertical flip the transform and place it on
    # the other side.
    # rho padding simply repeats the top/bottom rows into the padded region
    # negative padding values will result in a crop (remove the padding).
    shp = radon.shape
    if (tPad==0)&(rPad == 0):
      return radon


    if (mirrorTheta == True)&(tPad > 0):
      rdnP = np.zeros((shp[0],shp[1],shp[2] + 2 * tPad),dtype=np.float32)
      rdnP[:,:,tPad:-tPad] = radon
        # now pad out the theta dimension with the flipped-wrapped radon -- accounting for radon periodicity
      rdnP[:,:,0:tPad] = np.flip(rdnP[:,:,-2 *tPad:-tPad],axis=1)
      rdnP[:,:,-tPad:] = np.flip(rdnP[:,:,tPad:2 * tPad],axis=1)
    else:
      if tPad > 0:
        rdnP = np.pad(radon,((0,),(0,),(tPad,)),mode='edge')
      elif tPad < 0:
        rdnP = radon[:,:,-tPad:tPad]
      else:
        rdnP = radon

    if rPad > 0:
      rdnP =  np.pad(rdnP,((0,),(rPad,),(0,)),mode='edge')
    elif rPad < 0:
      rdnP = rdnP[:,-rPad:rPad, :]

    return rdnP



  def rdn_conv(self, radonIn):
    tic = timer()
    shp = radonIn.shape
    if len(shp) == 2:
      radon = radonIn.reshape(shp[0],shp[1],1)
      shp = radon.shape
    else:
      radon = radonIn
    shprdn = radon.shape
    #rdnNormP = self.radonPad(radon,rPad=0,tPad=self.peakPad[1],mirrorTheta=True)
    if self.padding[1] > 0:
      radon[:,0:self.padding[1],:] = np.flip(radon[:,-2 * self.padding[1]:-self.padding[1],:],axis=0)
      radon[:,-self.padding[1]:,:] = np.flip(radon[:,self.padding[1]:2 * self.padding[1],:],axis=0)
    # pad again for doing the convolution
    #rdnNormP = self.radonPad(rdnNormP,rPad=self.peakPad[0],tPad=self.peakPad[1],mirrorTheta=False)
    if self.padding[0] > 0:
      radon[0:self.padding[0], :,:] = radon[self.padding[0],:,:].reshape(1,shp[1], shp[2])
      radon[-self.padding[0]:, :,:] = radon[-self.padding[0]-1, :,:].reshape(1, shp[1],shp[2])


    rdnConv = np.zeros_like(radon)

    for i in range(shp[2]):
      rdnConv[:,:,i] = -1.0 * scipyndim.gaussian_filter(np.squeeze(radon[:,:,i]),[self.rSigma,self.tSigma],order=[2,0])

    #print(rdnConv.min(),rdnConv.max())
    mns = (rdnConv[self.padding[0]:shprdn[1]-self.padding[0],self.padding[1]:shprdn[1]-self.padding[1],:]).min(axis=0).min(axis=0)

    rdnConv -= mns.reshape((1,1, shp[2]))
    rdnConv = rdnConv.clip(min=0.0)

    return rdnConv

  def rdn_local_max(self, rdn, clparams=None, rdn_gpu=None, use_gpu=False):

    shp = rdn.shape
    # find the local max
    lMaxK = (self.peakPad[0],self.peakPad[1],1)

    lMaxRdn = scipyndim.grey_dilation(rdn,size=lMaxK)
    #lMaxRdn[:,:,0:self.peakPad[1]] = 0
    #lMaxRdn[:,:,-self.peakPad[1]:] = 0
    #location of the max is where the local max is equal to the original.
    lMaxRdn = lMaxRdn == rdn

    #rhoMaskTrim = np.int32((shp[0] - 2 * self.padding[0]) * self.rhoMaskFrac + self.padding[0])
    #lMaxRdn[0:rhoMaskTrim,:,:] = 0
    #lMaxRdn[-rhoMaskTrim:,:,:] = 0
    #lMaxRdn[:,0:self.padding[1],:] = 0
    #lMaxRdn[:,-self.padding[1]:,:] = 0
    #print("Traditional:",timer() - tic)
    maskrnd = np.zeros((self.nRho + 2 * self.padding[0], self.nTheta + 2 * self.padding[1],1), dtype=np.ubyte)
    maskrnd[self.padding[0]:-self.padding[0], self.padding[1]:-self.padding[1],0] = self.rdnmask.astype(np.ubyte)
    lMaxRdn *= maskrnd.astype(bool)
    return lMaxRdn


  def band_label(self,nPats,rdnConvIn,rdnNormIn,lMaxRdnIn):
    bandData = np.zeros((nPats,self.nBands),dtype=self.dataType)


    bdat = self.band_label_numba(
      np.int64(self.nBands),
      np.int64(nPats),
      np.int64(self.nRho),
      np.int64(self.nTheta),
      rdnConvIn,
      rdnConvIn,
      lMaxRdnIn
    )

    bandData['max']    = bdat[0][0:nPats, :]
    bandData['avemax'] = bdat[1][0:nPats, :]
    bandData['maxloc'] = bdat[2][0:nPats, :, :]
    bandData['aveloc'] = bdat[3][0:nPats, :, :]
    bandData['valid']  = bdat[4][0:nPats, :]
    bandData['width']  = bdat[5][0:nPats, :]
    bandData['maxloc'] -= self.padding.reshape(1, 1, 2)
    bandData['aveloc'] -= self.padding.reshape(1, 1, 2)

    return bandData

  @staticmethod
  @numba.jit(nopython=True,fastmath=True,cache=True,parallel=False)
  def band_label_numba(nBands,nPats,nRho,nTheta,rdnConv,rdnPad,lMaxRdn):
    nB = np.int64(nBands)
    nP = np.int64(nPats)

    shp = rdnPad.shape

    bandData_max = np.zeros((nP,nB),dtype=np.float32) - 2.0e6  # max of the convolved peak value
    bandData_avemax = np.zeros((nP,nB),
                               dtype=np.float32) - 2.0e6  # mean of the nearest neighborhood values around the max
    bandData_valid = np.zeros((nP, nB), dtype=np.int8)
    bandData_maxloc = np.zeros((nP,nB,2),dtype=np.float32)  # location of the max within the radon transform
    bandData_aveloc = np.zeros((nP,nB,2),
                               dtype=np.float32)  # location of the max based on the nearest neighbor interpolation
    bandData_width = np.zeros((nP,nB),dtype=np.float32) # a metric of the band width

    #nnc = np.array([-2,-1,0,1,2,-2,-1,0,1,2,-2,-1,0,1,2],dtype=np.float32)
    #nnr = np.array([-1,-1,-1,-1,-1,0,0,0,0,0,1,1,1,1,1],dtype=np.float32)
    #nnN = numba.float32(15)
    nnc = np.array([-1,0,1,-1,0,1,-1,0,1],dtype=np.float32)
    nnr = np.array([-1, -1, -1, 0, 0, 0, 1, 1, 1], dtype=np.float32)
    nnN = numba.float32(9)
    for q in range(nPats):
      # rdnConv_q = np.copy(rdnConv[:,:,q])
      # rdnPad_q = np.copy(rdnPad[:,:,q])
      # lMaxRdn_q = np.copy(lMaxRdn[:,:,q])
      # peakLoc = np.nonzero((lMaxRdn_q == rdnPad_q) & (rdnPad_q > 1.0e-6))
      peakLoc = lMaxRdn[:,:,q].nonzero()
      indx1D = peakLoc[1] + peakLoc[0] * shp[1]
      temp = (rdnConv[:,:,q].ravel())[indx1D]
      srt = np.argsort(temp)
      nBq = nB if (len(srt) > nB) else len(srt)
      for i in numba.prange(nBq):
        r = np.int32(peakLoc[0][srt[-1 - i]])
        c = np.int32(peakLoc[1][srt[-1 - i]])
        bandData_maxloc[q,i,:] = np.array([r,c])
        bandData_max[q,i] = rdnPad[r,c,q]
        bandData_width[q, i] = 1.0 / (bandData_max[q,i] - 0.5* (rdnPad[r+1, c, q] + rdnPad[r-1, c, q]) + 1.0e-12)

        #center of mass peak localization
        #nn = rdnConv[r - 1:r + 2,c - 2:c + 3,q].ravel()
        #sumnn = (np.sum(nn) + 1.e-12)
        #nn /= sumnn
        #bandData_avemax[q,i] = sumnn / nnN
        #rnn = np.sum(nn * (np.float32(r) + nnr))
        #cnn = np.sum(nn * (np.float32(c) + nnc))

        # taylor expansion quadratic
        nn = rdnConv[r - 1:r + 2,c - 1:c + 2,q].copy()
        sumnn = (np.sum(nn) + 1.e-12)
        nn /= sumnn
        bandData_avemax[q,i] = sumnn / nnN
        # rnn = np.sum(nn * (np.float32(r) + nnr))
        # cnn = np.sum(nn * (np.float32(c) + nnc))
        dx  = 0.5*(nn[1,2] - nn[1,0])
        dy  = 0.5*(nn[2,1] - nn[0,1])
        dxx = nn[1,2] + nn[1,0] - 2 * nn[1,1]
        dyy = nn[2,1] + nn[0,1] - 2 * nn[1,1]
        dxy = 0.25*(nn[2,2] - nn [0,2] - nn[2,0] + nn[0,0])
        #det = 1.0 / (dxx * dyy - dxy * dxy)
        det = (dxx * dyy - dxy * dxy)
        det = det if np.fabs(det) > 1e-12 else 1.0e-12
        det = 1.0/det
        dc =  (dyy * dx - dxy * dy) * det
        rc = (dxx * dy - dxy * dx) * det
        dc = max(-1.0, dc) ; rc = max(-1.0, rc)
        dc = min(1.0, dc) ;  rc = min(1.0, rc)
        cnn = c - dc
        rnn = r - rc
        bandData_aveloc[q,i,:] = np.array([rnn,cnn])
        bandData_valid[q,i] = 1

    return bandData_max,bandData_avemax,bandData_maxloc,bandData_aveloc, bandData_valid, bandData_width
  def _display_radon_pattern(self, rdnConvarray, bandData, patterns):
    if len(rdnConvarray.shape) == 3:
      im2show = rdnConvarray[self.padding[0]:-self.padding[0], self.padding[1]:-self.padding[1], -1]
    else:
      im2show = rdnConvarray[self.padding[0]:-self.padding[0], self.padding[1]:-self.padding[1]]

    im2show *= self.rdnmask
    #rhoMaskTrim = np.int32(im2show.shape[0] * self.rhoMaskFrac)
    #mean = np.mean(im2show[rhoMaskTrim:-rhoMaskTrim, 1:-2])
    #stdv = np.std(im2show[rhoMaskTrim:-rhoMaskTrim, 1:-2])
    mean = np.mean(im2show, where=(self.rdnmask >0))
    stdv = np.std(im2show, where=(self.rdnmask >0))


    im2show -= mean
    im2show /= stdv
    im2show = im2show.clip(-4, None)
    im2show += 6
    #im2show[0:rhoMaskTrim, :] = 0
    #im2show[-rhoMaskTrim:, :] = 0

    im2show = np.fliplr(im2show)
    fig = plt.figure(figsize=(12, 4))
    subrdn = fig.add_subplot(121, xlim=(0, 180), ylim=(-self.rhoMax, self.rhoMax))
    subrdn.imshow(
      im2show,
      cmap='gray',
      extent=[0, 180, -self.rhoMax, self.rhoMax],
      interpolation='none',
      zorder=1,
      aspect='auto'
    )
    width = bandData['width'][-1, :]
    width /= width.min()
    width *= 2.0
    xplt = np.squeeze(
      180.0 - np.interp(bandData['aveloc'][-1, :, 1] + 0.5, np.arange(self.radonPlan.nTheta), self.radonPlan.theta))
    yplt = np.squeeze(
      -1.0 * np.interp(bandData['aveloc'][-1, :, 0] - 0.5, np.arange(self.radonPlan.nRho), self.radonPlan.rho))

    subrdn.scatter(y=yplt, x=xplt, c='r', s=width, zorder=2)

    for pt in range(self.nBands):
      subrdn.annotate(str(pt + 1), np.squeeze([xplt[pt] + 4, yplt[pt]]), color='yellow')
    # subrdn.xlim(0,180)
    # subrdn.ylim(-self.rhoMax, self.rhoMax)
    subpat = fig.add_subplot(122)
    pat1 = patterns[-1, :, :].copy().squeeze().astype(float)
    minpat = pat1.min()
    pdim = pat1.shape
    pat1 = np.concatenate((pat1.flatten(), [minpat]))
    patmask = np.copy(self.radonPlan.mask).astype(int)

    #patdisplay = np.zeros(patmask.size)
    patdisplay = pat1[[self.radonPlan.maskindex.flatten().clip(-1,).astype(int)]]
    patdisplay = patdisplay.reshape(self.radonPlan.maskindex.shape)

    patdisplay *= (patmask != 0).astype(int)
    patdisplay += minpat*(patmask ==0).astype(float)

    subpat.imshow(patdisplay, cmap='gray')
  def _circmask(self, im):
    nx = im.shape[-1]
    ny = im.shape[-2]
    # plt.imshow(back)
    x = np.arange(nx, dtype=float)
    x = (np.broadcast_to(x.reshape(1, nx), (ny, nx))).ravel()
    y = np.arange(ny, dtype=float)
    y = (np.broadcast_to(y, (nx, ny)).T).ravel()

      # make a circular mask - even if not EDAX, this should work OK.
    cx = (np.arange(nx) - nx * 0.5) ** 2
    cy = (np.arange(ny) - ny * 0.5) ** 2
    cmask = np.sqrt(np.broadcast_to(cx.reshape(1, nx), (ny, nx)) + np.broadcast_to(cy, (nx, ny)).T) #< (ny * 0.49)
    return cmask
def getopenclparam(**kwargs): # dummy function to maintain compatability with openCL version
  return None





# This software was developed by employees of the US Naval Research Laboratory (NRL), an
# agency of the Federal Government. Pursuant to title 17 section 105 of the United States
# Code, works of NRL employees are not subject to copyright protection, and this software
# is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
# responsibility whatsoever for its use by other parties, and makes no guarantees,
# expressed or implied, about its quality, reliability, or any other characteristic. We
# would appreciate acknowledgment if the software is used. To the extent that NRL may hold
# copyright in countries other than the United States, you are hereby granted the
# non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
# works and distribute this software, in any medium, or authorize others to do so on your
# behalf, on a royalty-free basis throughout the world. You may improve, modify, and
# create derivative works of the software or any portion of the software, and you may copy
# and distribute such modifications or works. Modified works should carry a notice stating
# that you changed the software and should note the date and nature of any such change.
# Please explicitly acknowledge the US Naval Research Laboratory as the original source.
# This software can be redistributed and/or modified freely provided that any derivative
# works bear some notice that they are derived from it, and any modified versions bear
# some notice that they have been modified.
#
# Author: David Rowenhorst;
# The US Naval Research Laboratory Date: 22 May 2024

# For more info see
# Patrick T. Brewick, Stuart I. Wright, David J. Rowenhorst. Ultramicroscopy, 200:50–61, May 2019.

"""Non-local pattern averaging and re-indexing (NLPAR)."""



import os, sys, platform
import logging
from timeit import default_timer as timer
import numpy as np
import pyopencl as cl

import ray


from pyebsdindex.opencl import openclparam, nlpar_cl
from time import time as timer

RAYIPADDRESS = '127.0.0.1'
OSPLATFORM  = platform.system()
#if OSPLATFORM  == 'Darwin':
#    RAYIPADDRESS = '0.0.0.0'  # the localhost address does not work on macOS when on a VPN


class NLPAR(nlpar_cl.NLPAR):
  def __init__( self,filename=None, **kwargs):
    nlpar_cl.NLPAR.__init__(self,filename=filename, **kwargs)
    self.useCPU = False

  def calcnlpar(self, **kwargs):
    return self.calcnlpar_clray(**kwargs)

  def calcsigma(self, nn=1, saturation_protect=True, automask=True, return_nndist=False, **kwargs):
    if self.sigmann > 7:
      print("Sigma optimization search limited to a search radius <= 7")
      print("The search radius has been clipped to 7")
      nn = 7
      self.sigmann = nn

    sig = self.calcsigma_clray(nn=nn,
                            saturation_protect=saturation_protect,
                            automask=automask, **kwargs)
    if return_nndist == True:
      return sig
    else:
      return sig[0]


  def calcnlpar_clsq(self, **kwargs):
    return nlpar_cl.NLPAR.calcnlpar_cl(self, **kwargs)

  def calcsigma_clsq(self, **kwargs):
    return nlpar_cl.NLPAR.calcsigma_cl(self, **kwargs)

  def calcsigma_clray(self, nn=1, saturation_protect=True, automask=True, normalize_d=False,
                      gpu_id = None, verbose=2, **kwargs):
    self.patternfile = self.getinfileobj()
    self.sigmann = nn

    if self.sigmann > 7:
      print("Sigma optimization search limited to a search radius <= 7")
      print("The search radius has been clipped to 7")
      nn = 7
      self.sigmann = nn

    if gpu_id is None:
      clparams = openclparam.OpenClParam()
      clparams.get_gpu()

      target_mem = 0
      gpu_id = 0
      count = 0

      for gpu in clparams.gpu:
        gmem = gpu.max_mem_alloc_size
        if target_mem < gmem:
          gpu_id = count
          target_mem = gmem
        count += 1
    else:
      clparams = openclparam.OpenClParam()
      clparams.get_gpu()
      gpu_id = min(len(clparams.gpu)-1, gpu_id)

    cudavis = ''
    for cdgpu in range(len(clparams.gpu)):
      cudavis += str(cdgpu) + ','

    #print(gpu_id)
    ngpuwrker = 6
    clparams.get_context(gpu_id=gpu_id, kfile = 'clnlpar.cl')
    clparams.get_queue()
    if clparams.gpu[gpu_id].host_unified_memory:
      return nlpar_cl.NLPAR.calcsigma_cl(self, nn=nn, saturation_protect=saturation_protect,
                                         automask=automask,
                                         normalize_d=normalize_d,
                                         gpu_id=gpu_id, **kwargs)

    target_mem = clparams.gpu[gpu_id].max_mem_alloc_size // 2
    max_mem = clparams.gpu[gpu_id].global_mem_size * 0.5
    if target_mem * ngpuwrker > max_mem:
      #print('revisemem:')
      target_mem = max_mem / ngpuwrker

    patternfile = self.getinfileobj()

    nrows = np.int64(self.nrows)  # np.uint64(patternfile.nRows)
    ncols = np.int64(self.ncols)  # np.uint64(patternfile.nCols)

    pwidth = np.uint64(patternfile.patternW)
    pheight = np.uint64(patternfile.patternH)
    npat_point = int(pwidth * pheight)

    if (automask is True) and (self.mask is None):
      self.mask = (self.automask(pheight, pwidth))
    if self.mask is None:
      self.mask = np.ones((pheight, pwidth), dtype=np.uint8)

    mask = self.mask.astype(np.float32)

    chunks = self._calcchunks([pwidth, pheight], ncols, nrows, target_bytes=target_mem,
                              col_overlap=nn, row_overlap=nn)


    jobqueue = []

    for rowchunk in range(chunks[1]):
      rstart = chunks[3][rowchunk, 0]
      rend = chunks[3][rowchunk, 1]
      nrowchunk = rend - rstart

      rstartcalc = nn if (rowchunk > 0) else 0
      rendcalc = nrowchunk - nn if (rowchunk < (chunks[1] - 1)) else nrowchunk
      nrowcalc = np.int64(rendcalc - rstartcalc)

      for colchunk in range(chunks[0]):
        cstart = chunks[2][colchunk, 0]
        cend = chunks[2][colchunk, 1]
        ncolchunk = cend - cstart

        cstartcalc = nn if (colchunk > 0) else 0
        cendcalc = ncolchunk - nn if (colchunk < (chunks[0] - 1)) else ncolchunk
        ncolcalc = np.int64(cendcalc - cstartcalc)

        jobqueue.append(NLPARGPUJob([colchunk, rowchunk], \
                                    [cstart, cend, rstart, rend], \
                                    [cstartcalc, cendcalc, rstartcalc, rendcalc]))


    # wrker = NLPARGPUWorker(actorid=1, gpu_id=gpu_id, cudavis=cudavis)
    # job = jobqueue[0]
    #
    # data = wrker.runsigma_chunk(job, nlparobj=self, saturation_protect=saturation_protect)
    #
    # return data

    ngpu_per_wrker = float(1.0 / ngpuwrker)

    # if verbose >=1:
    #   print("lambda:", self.lam, "search radius:", self.searchradius, "dthresh:", self.dthresh)

    ray.shutdown()
    rayclust = ray.init(
      num_cpus=int(ngpuwrker),
      num_gpus=1,
      # dashboard_host = RAYIPADDRESS,
      _node_ip_address=RAYIPADDRESS,  # "0.0.0.0",
      runtime_env={"env_vars":
                     {"PYTHONPATH": os.path.dirname(os.path.dirname(os.path.dirname(__file__))),
                      }},
      logging_level=logging.WARNING, )  # Supress INFO messages from ray.

    nlpar_remote = ray.put(self)

    nnn = int((2 * nn + 1) ** 2)
    sigma = np.zeros((nrows, ncols), dtype=np.float32) + 1e18
    dist = np.zeros((nrows, ncols, nnn), dtype=np.float32)
    countnn = np.zeros((nrows, ncols, nnn), dtype=np.float32)

    idlewrker = []
    busywrker = []
    tasks = []

    for w in range(ngpuwrker):
      idlewrker.append(NLPARGPUWorker.options(num_cpus=float(0.99), num_gpus=ngpu_per_wrker).remote(
        actorid=w, gpu_id=gpu_id, cudavis=cudavis))

    njobs = len(jobqueue)
    ndone = 0

    while ndone < njobs:
      if len(jobqueue) > 0:
        if len(idlewrker) > 0:
          wrker = idlewrker.pop()
          job = jobqueue.pop()

          tasks.append(wrker.runsigma_chunk.remote(job, nlparobj=nlpar_remote, saturation_protect=saturation_protect))
          busywrker.append(wrker)
      if len(tasks) > 0:
        donetasks, stillbusy = ray.wait(tasks, num_returns=len(busywrker), timeout=0.1)

        for tsk in donetasks:
          indx = tasks.index(tsk)
          message, job, newdata = ray.get(tsk)
          if message == 'Done':

            sigmachunk = newdata[0]
            distchunk = newdata[1]
            countchunk = newdata[2]
            cstart = job.cstart
            cend = job.cend
            ncolchunk = job.ncolchunk
            rstart = job.rstart
            rend = job.rend
            nrowchunk = job.nrowchunk

            rstartcalc = job.rstartcalc
            cstartcalc = job.cstartcalc
            nrowcalc = job.nrowcalc
            ncolcalc = job.ncolcalc

            countnn[rstart:rend, cstart:cend] = countchunk[0:int(ncolchunk * nrowchunk), :].reshape(nrowchunk,
                                                                                        ncolchunk, nnn)
            dist[rstart:rend, cstart:cend] = distchunk[0:int(ncolchunk * nrowchunk), :].reshape(nrowchunk, ncolchunk,
                                                                                                nnn)
            sigma[rstart:rend, cstart:cend] = np.minimum(sigma[rstart:rend, cstart:cend], sigmachunk)

            idlewrker.append(busywrker.pop(indx))
            tasks.remove(tsk)
            ndone += 1
          if verbose >= 2:
            print("tiles complete: ", ndone,"/", njobs,sep='', end='\r'  )
    if verbose >= 2:
      print('\n', end='')
    return sigma, dist, countnn

  def _sigmachunkcalc_cl(self, data, calclim, clparams=None, saturation_protect=True):
    nn = self.sigmann

    data = np.ascontiguousarray(data)
    ctx = clparams.ctx
    prg = clparams.prg
    clparams.get_queue()


    mf = clparams.memflags
    clvectlen = 16

    cstart = calclim.cstart
    cend = calclim.cend
    ncolchunk = calclim.ncolchunk
    rstart = calclim.rstart
    rend = calclim.rend
    nrowchunk = calclim.nrowchunk

    rstartcalc = calclim.rstartcalc
    cstartcalc = calclim.cstartcalc
    nrowcalc = calclim.nrowcalc
    ncolcalc = calclim.ncolcalc

    pheight = data.shape[1]
    pwidth = data.shape[2]
    npat_point = pheight*pwidth

    mask = self.mask

    npad = clvectlen * int(np.ceil(mask.size/clvectlen))
    maskpad = np.zeros((npad) , np.float32)
    maskpad[0:mask.size] = mask.reshape(-1).astype(np.float32)
    mask_gpu = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=maskpad)

    npatsteps = int(maskpad.size/clvectlen)

    #return chunks, chunksize
    mxchunk = data.shape[0]

    npadmx = clvectlen * int(np.ceil(mxchunk*npat_point/ clvectlen))

    datapad_gpu = cl.Buffer(ctx, mf.READ_WRITE, size=int(npadmx) * int(4))

    nnn = int((2 * nn + 1) ** 2)

    #dist_local = cl.LocalMemory(nnn*npadmx*4)
    dist_local = cl.Buffer(ctx, mf.READ_WRITE, size=int(mxchunk*nnn*4))
    distchunk = np.zeros((mxchunk, nnn), dtype=np.float32)
    #count_local = cl.LocalMemory(nnn*npadmx*4)
    count_local = cl.Buffer(ctx, mf.READ_WRITE, size=int(mxchunk * nnn * 4))
    countchunk = np.zeros((mxchunk, nnn), dtype=np.float32)

    mxval = data.max()
    if saturation_protect == False:
      mxval += 1.0
    else:
      mxval *= 0.9961

    evnt = cl.enqueue_fill_buffer(clparams.queue, datapad_gpu, np.float32(mxval+10), 0,int(4*npadmx))

    szdata = data.size
    npad = clvectlen * int(np.ceil(szdata / clvectlen))
    tic = timer()
    #datapad = np.zeros((npad), dtype=np.float32) + np.float32(mxval + 10)
    #datapad[0:szdata] = data.reshape(-1)
    data_gpu = cl.Buffer(ctx,mf.READ_ONLY | mf.COPY_HOST_PTR,hostbuf=data)

    if data.dtype.type is np.float32:
      prg.nlloadpat32flt(clparams.queue, (np.uint64(data.size),), None, data_gpu, datapad_gpu, wait_for=[evnt])
    if data.dtype.type is np.ubyte:
      prg.nlloadpat8bit(clparams.queue, (np.uint64(data.size),), None, data_gpu, datapad_gpu, wait_for=[evnt])
    if data.dtype.type is np.uint16:
      prg.nlloadpat16bit(clparams.queue, (np.uint64(data.size),), None, data_gpu, datapad_gpu, wait_for=[evnt])
    toc = timer()
    #print(toc - tic)

    sigmachunk = np.zeros((nrowchunk, ncolchunk ), dtype=np.float32)


    sigmachunk_gpu =  cl.Buffer(ctx, mf.WRITE_ONLY, size=sigmachunk.nbytes)
    cl.enqueue_barrier(clparams.queue)
    prg.calcsigma(clparams.queue, (np.uint32(ncolchunk), np.uint32(nrowchunk)), None,
                           datapad_gpu, mask_gpu,sigmachunk_gpu,
                           dist_local, count_local,
                           np.int64(nn), np.int64(npatsteps), np.int64(npat_point),
                           np.float32(mxval) )

    cl.enqueue_barrier(clparams.queue)
    prg.normd(clparams.queue, (np.uint32(ncolchunk), np.uint32(nrowchunk)), None,
                    sigmachunk_gpu,
                    count_local, dist_local,
                    np.int64(nn))
    clparams.queue.finish()

    cl.enqueue_copy(clparams.queue, distchunk, dist_local, is_blocking=False)
    cl.enqueue_copy(clparams.queue, countchunk, count_local,  is_blocking=False)
    cl.enqueue_copy(clparams.queue, sigmachunk, sigmachunk_gpu,  is_blocking=True)

    #sigmachunk_gpu.release()

    clparams.queue.finish()

    dist_local.release()
    count_local.release()
    datapad_gpu.release()
    clparams.queue.flush()
    clparams.queue = None
    #self.sigma = sigma
    return sigmachunk, distchunk, countchunk



  def calcnlpar_clray(self, searchradius=None, lam = None, dthresh = None, saturation_protect=True, automask=True,
                filename=None, fileout=None, reset_sigma=False, backsub = False, rescale = False,
                verbose = 2, gpu_id = None, **kwargs):

    if lam is not None:
      self.lam = lam

    self.saturation_protect = saturation_protect

    if dthresh is not None:
      self.dthresh = dthresh
    if self.dthresh is None:
      self.dthresh = 0.0

    if searchradius is not None:
      self.searchradius = searchradius

    if self.searchradius > 10:
      print("NLPAR on GPU is limited to a search radius <= 10")
      print("The search radius has been clipped to 10")
      searchradius = 10
      self.searchradius = searchradius

    lam = np.float32(self.lam)
    dthresh = np.float32(self.dthresh)
    sr = np.int64(self.searchradius)

    if filename is not None:
      self.setfile(filepath=filename)

    self.patternfile = self.getinfileobj()
    self.setoutfile(self.patternfile, filepath=fileout)
    self.patternfileout = self.getoutfileobj()

    nrows = np.int64(self.nrows)  # np.uint64(patternfile.nRows)
    ncols = np.int64(self.ncols)  # np.uint64(patternfile.nCols)

    pwidth = np.uint64(self.patternfile.patternW)
    pheight = np.uint64(self.patternfile.patternH)
    npat_point = int(pwidth * pheight)

    if reset_sigma:
      self.sigma = None

    if np.asarray(self.sigma).size == 1 and (self.sigma is not None):
      tmp = np.asarray(self.sigma)[0]
      self.sigma = np.zeros((nrows, ncols), dtype=np.float32) + tmp
      calcsigma = False

    if self.sigma is not None:
      shpsigma = np.asarray(self.sigma).shape
      if (shpsigma[0] != nrows) and (shpsigma[1] != ncols):
        self.sigma = None

    if self.sigma is None:
      self.sigma = self.calcsigma_cl(nn=1, saturation_protect=saturation_protect, automask=automask, gpu_id=gpu_id)[0]

    sigma = np.asarray(self.sigma).astype(np.float32)



    if (automask is True) and (self.mask is None):
      self.mask = (self.automask(pheight, pwidth))
    if self.mask is None:
      self.mask = np.ones((pheight, pwidth), dtype=np.uint8)

    scalemethod = 'clip'
    self.rescale = False
    if rescale == True:
      self.rescale = True
      if np.issubdtype(self.patternfileout.filedatatype, np.integer):
        mxval = np.iinfo(self.patternfileout.filedatatype).max
        scalemethod = 'fullscale'
      else:  # not int, so no rescale.
        self.rescale = False

    ngpuwrker = 6
    clparams = openclparam.OpenClParam()
    clparams.get_gpu()
    if gpu_id is None:
      target_mem = 0
      gpu_id = 0
      count = 0

      for gpu in clparams.gpu:
        gmem = gpu.max_mem_alloc_size
        if target_mem < gmem:
          gpu_id = count
          target_mem = gmem
        count += 1
    else:
      gpu_id = min(len(clparams.gpu)-1, gpu_id)
    cudavis = ''
    for cdgpu in range(len(clparams.gpu)):
        cudavis += str(cdgpu) + ','

    # print(gpu_id)
    clparams.get_context(gpu_id=gpu_id, kfile = 'clnlpar.cl')
    clparams.get_queue()
    if clparams.gpu[gpu_id].host_unified_memory:
      return nlpar_cl.NLPAR.calcnlpar_cl(self, saturation_protect=saturation_protect,
                                         automask=automask,
                                         filename=filename,
                                         fileout=fileout,
                                         reset_sigma=reset_sigma,
                                         backsub = backsub,
                                         rescale = rescale,
                                         gpu_id= gpu_id)

    target_mem = clparams.gpu[gpu_id].max_mem_alloc_size//6
    max_mem = clparams.gpu[gpu_id].global_mem_size*0.4
    if target_mem*ngpuwrker > max_mem:
      target_mem = max_mem/ngpuwrker
    #print(target_mem/1.0e9)

    chunks = self._calcchunks([pwidth, pheight], ncols, nrows, target_bytes=target_mem,
                              col_overlap=sr, row_overlap=sr)

    nnn = int((2 * sr + 1) ** 2)
    jobqueue = []


    for rowchunk in range(chunks[1]):
        rstart = chunks[3][rowchunk, 0]
        rend = chunks[3][rowchunk, 1]
        nrowchunk = rend - rstart

        rstartcalc = sr if (rowchunk > 0) else 0
        rendcalc = nrowchunk - sr if (rowchunk < (chunks[1] - 1)) else nrowchunk
        nrowcalc = np.int64(rendcalc - rstartcalc)

        for colchunk in range(chunks[0]):
            cstart = chunks[2][colchunk, 0]
            cend = chunks[2][colchunk, 1]
            ncolchunk = cend - cstart

            cstartcalc = sr if (colchunk > 0) else 0
            cendcalc = ncolchunk - sr if (colchunk < (chunks[0] - 1)) else ncolchunk
            ncolcalc = np.int64(cendcalc - cstartcalc)

            jobqueue.append( NLPARGPUJob([colchunk, rowchunk],\
                    [cstart,cend, rstart, rend],\
                  [cstartcalc,cendcalc, rstartcalc, rendcalc ]))


    if verbose >=1:
      print("lambda:", lam, "search radius:", sr, "dthresh:", dthresh)
    ngpu_per_wrker = float(1.0/ngpuwrker)
    ray.shutdown()

    rayclust = ray.init(
      num_cpus=int(ngpuwrker),
      num_gpus=1,
      # dashboard_host = RAYIPADDRESS,
      _node_ip_address=RAYIPADDRESS,  # "0.0.0.0",
      runtime_env={"env_vars":
                     {"PYTHONPATH": os.path.dirname(os.path.dirname(os.path.dirname(__file__))),
                      }},
      logging_level=logging.WARNING,)  # Supress INFO messages from ray.

    nlpar_remote = ray.put(self)

    idlewrker = []
    busywrker = []
    tasks = []

    for w in range(ngpuwrker):
        idlewrker.append(NLPARGPUWorker.options(num_cpus=float(0.99), num_gpus=ngpu_per_wrker).remote(
                actorid=w, gpu_id=gpu_id, cudavis=cudavis))

    njobs = len(jobqueue)
    ndone = 0
    while ndone < njobs:
        if len(jobqueue) > 0:
            if len(idlewrker) > 0:
                wrker = idlewrker.pop()
                job = jobqueue.pop(0)

                tasks.append(wrker.runnlpar_chunk.remote(job, nlparobj=nlpar_remote))
                busywrker.append(wrker)
        if len(tasks) > 0:
            donetasks, stillbusy = ray.wait(tasks, num_returns=len(busywrker), timeout=0.1)

            for tsk in donetasks:
                indx = tasks.index(tsk)
                message, job, newdata = ray.get(tsk)
                if message == 'Done':
                  idlewrker.append(busywrker.pop(indx))
                  tasks.remove(tsk)
                  ndone += 1
                  if verbose >= 2:
                    print("tiles complete: ", ndone, "/", njobs, sep='', end='\r')
                else: #An error has occured ... hopefully just need a re-process.
                  jobqueue.append(job)
                  print(message)

    if verbose >= 2:
      print('\n', end='')
    return str(self.patternfileout.filepath)

  def _nlparchunkcalc_cl(self, data, calclim, clparams=None):
    data = np.ascontiguousarray(data)
    ctx = clparams.ctx
    prg = clparams.prg
    clparams.get_queue()


    mf = clparams.memflags
    clvectlen = 16

    cstart = calclim.cstart
    cend = calclim.cend
    ncolchunk = calclim.ncolchunk
    rstart = calclim.rstart
    rend = calclim.rend
    nrowchunk = calclim.nrowchunk

    rstartcalc = calclim.rstartcalc
    cstartcalc = calclim.cstartcalc
    nrowcalc = calclim.nrowcalc
    ncolcalc = calclim.ncolcalc

    pheight = data.shape[1]
    pwidth = data.shape[2]



    lam = np.float32(self.lam)
    sr = np.int64(self.searchradius)
    nnn = int((2 * sr + 1) ** 2)
    dthresh = np.float32(self.dthresh)
    #print(chunks[2], chunks[3])
    #print(lam, sr, dthresh)


    # precalculate some needed arrays for the GPU
    mxval = data.max()
    if self.saturation_protect == False:
      mxval += 1.0
    else:
      mxval *= 0.9961

    shpdata = data.shape
    npat_point = int(shpdata[-1]*shpdata[-2])
    npat = shpdata[0]

    npadmx = clvectlen * int(np.ceil(float(npat)*npat_point/ clvectlen))
    data_gpu = cl.Buffer(ctx, mf.READ_ONLY| mf.COPY_HOST_PTR, hostbuf=data)
    datapad_gpu = cl.Buffer(ctx, mf.READ_WRITE, size=int(npadmx) * int(4))
    datapadout_gpu = cl.Buffer(ctx, mf.READ_WRITE, size=int(npadmx) * int(4))

    fill1 = cl.enqueue_fill_buffer(clparams.queue, datapad_gpu, np.float32(mxval + 10), 0, int(4 * npadmx))
    fill2 = cl.enqueue_fill_buffer(clparams.queue, datapadout_gpu, np.float32(0.0), 0, int(4 * npadmx))


    mask = self.mask.astype(np.float32)

    npad = clvectlen * int(np.ceil(mask.size / clvectlen))
    maskpad = np.zeros((npad), np.float32) - 1  # negative numbers will indicate a clvector overflow.
    maskpad[0:mask.size] = mask.reshape(-1).astype(np.float32)
    mask_gpu = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=maskpad)
    npatsteps = int(maskpad.size / clvectlen)

    sigmachunk = np.ascontiguousarray(self.sigma[rstart:rend, cstart:cend].astype(np.float32))
    sigmachunk_gpu = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=sigmachunk)

    szdata = data.size
    cl.enqueue_barrier(clparams.queue)
    data_gpu = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=data)
    if data.dtype.type is np.float32:
      prg.nlloadpat32flt(clparams.queue, (np.uint64(data.size), 1), None, data_gpu, datapad_gpu)#, wait_for=[fill1,fill2])
    if data.dtype.type is np.ubyte:
      prg.nlloadpat8bit(clparams.queue, (np.uint64(data.size), 1), None, data_gpu, datapad_gpu)#, wait_for=[fill1,fill2])
    if data.dtype.type is np.uint16:
      prg.nlloadpat16bit(clparams.queue, (np.uint64(data.size), 1), None, data_gpu, datapad_gpu)#, wait_for=[fill1,fill2])

    calclim = np.array([cstartcalc, rstartcalc, ncolchunk, nrowchunk], dtype=np.int64)
    crlimits_gpu = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=calclim)
    cl.enqueue_barrier(clparams.queue)
    data_gpu.release()
    prg.calcnlpar(clparams.queue, (np.uint32(ncolcalc), np.uint32(nrowcalc)), None,
                  # prg.calcnlpar(clparams.queue, (1, 1), None,
                  datapad_gpu,
                  mask_gpu,
                  sigmachunk_gpu,
                  crlimits_gpu,
                  datapadout_gpu,
                  np.int64(sr),
                  np.int64(npatsteps),
                  np.int64(npat_point),
                  np.float32(mxval),
                  np.float32(1.0 / lam ** 2),
                  np.float32(dthresh))

    data = data.astype(np.float32)  # prepare to receive data back from GPU
    data.reshape(-1)[:] = 0.0
    data = data.reshape(nrowchunk, ncolchunk, pheight, pwidth)
    cl.enqueue_copy(clparams.queue, data, datapadout_gpu,  is_blocking=True)
    sigmachunk_gpu.release()
    clparams.queue.finish()
    if self.rescale == True:
      for i in range(data.shape[0]):
        temp = data[i, :, :]
        temp -= temp.min()
        temp *= np.float32(mxval) / temp.max()
        data[i, :, :] = temp
    data = data[rstartcalc: rstartcalc + nrowcalc, cstartcalc: cstartcalc + ncolcalc, :, :]
    data = data.reshape(nrowcalc * ncolcalc, pheight, pwidth)

    clparams.queue = None

    return data



@ray.remote
class NLPARGPUWorker:
  def __init__(self, actorid=0, gpu_id=None, cudavis = '0'):
      # sys.path.append(path.dirname(path.dirname(__file__)))  # do this to help Ray find the program files
      # import openclparam # do this to help Ray find the program files
      # device, context, queue, program, mf
      # self.dataout = None
      # self.indxstart = None
      # self.indxend = None
      # self.rate = None
      os.environ["CUDA_VISIBLE_DEVICES"] = cudavis
      self.actorID = actorid
      self.openCLParams = openclparam.OpenClParam()

      self.openCLParams.gpu_id = gpu_id
      self.openCLParams.get_context(gpu_id=gpu_id, kfile = 'clnlpar.cl')


          #elf.openCLParams = None

  def runsigma_chunk(self,gpujob, nlparobj=None, **kwargs):
    if gpujob is None:
        #time.sleep(0.001)
        return 'Bored', (None, None, None)
    try:
        # print(type(self.openCLParams.ctx))
        gpujob._starttime()
        #time.sleep(random.uniform(0, 1.0))

        #if self.openCLParams is not None:
        #    self.openCLParams.get_queue()
        data, xyloc = nlparobj.patternfile.read_data(patStartCount=[[gpujob.cstart, gpujob.rstart],
                                                                    [gpujob.ncolchunk, gpujob.nrowchunk]],
                                                                    convertToFloat=False, returnArrayOnly=True)

        newdata = nlparobj._sigmachunkcalc_cl(data, gpujob, clparams=self.openCLParams, **kwargs)

        if self.openCLParams.queue is not None:
            print("queue still here")
            self.openCLParams.queue.finish()
            self.openCLParams.queue = None

        gpujob._endtime()
        return 'Done', gpujob, newdata
    except Exception as e:
        print(e)
        gpujob.rate = None
        return "Error", gpujob, e


  def runnlpar_chunk(self, gpujob, nlparobj=None):

    if gpujob is None:
        #time.sleep(0.001)
        return 'Bored', (None, None, None)
    try:
        # print(type(self.openCLParams.ctx))
        gpujob._starttime()
        #time.sleep(random.uniform(0, 1.0))

        #if self.openCLParams is not None:
        #    self.openCLParams.get_queue()
        data, xyloc = nlparobj.patternfile.read_data(patStartCount=[[gpujob.cstart, gpujob.rstart],
                                                                    [gpujob.ncolchunk, gpujob.nrowchunk]],
                                                                    convertToFloat=False, returnArrayOnly=True)

        newdata = nlparobj._nlparchunkcalc_cl(data, gpujob, clparams=self.openCLParams)

        if self.openCLParams.queue is not None:
            print("queue still here")
            self.openCLParams.queue.finish()
            self.openCLParams.queue = None

        nlparobj.patternfileout.write_data(newpatterns=newdata,
                                       patStartCount=[[gpujob.cstart + gpujob.cstartcalc,
                                                       gpujob.rstart + gpujob.rstartcalc],
                                                      [gpujob.ncolcalc, gpujob.nrowcalc]],
                                       flt2int='clip', scalevalue=1.0)

        gpujob._endtime()
        return 'Done', gpujob, None
    except Exception as e:
        print(e)
        gpujob.rate = None
        return "Error", gpujob, e


class NLPARGPUJob:
  def __init__(self, jobid, chunk, calclim):
    self.jobid = jobid
    if self.jobid is not None:
      self.cstart = chunk[0]
      self.cend = chunk[1]
      self.ncolchunk = self.cend -  self.cstart
      self.rstart = chunk[2]
      self.rend = chunk[3]
      self.nrowchunk = self.rend - self.rstart
      self.cstartcalc = calclim[0]
      self.cendcalc = calclim[1]
      self.ncolcalc = self.cendcalc - self.cstartcalc
      self.rstartcalc = calclim[2]
      self.rendcalc = calclim[3]
      self.nrowcalc = self.rendcalc - self.rstartcalc
      self.starttime = 0.0
      self.endtime = 0.0
      self.extime = 0.0

  def _starttime(self):
    self.starttime = timer()

  def _endtime(self):
    self.endtime = timer()
    self.extime += self.endtime - self.starttime
    #self.rate = self.npat / (self.extime + 1e-12)






"""This software was developed by employees of the US Naval Research Laboratory (NRL), an
agency of the Federal Government. Pursuant to title 17 section 105 of the United States
Code, works of NRL employees are not subject to copyright protection, and this software
is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
responsibility whatsoever for its use by other parties, and makes no guarantees,
expressed or implied, about its quality, reliability, or any other characteristic. We
would appreciate acknowledgment if the software is used. To the extent that NRL may hold
copyright in countries other than the United States, you are hereby granted the
non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
works and distribute this software, in any medium, or authorize others to do so on your
behalf, on a royalty-free basis throughout the world. You may improve, modify, and
create derivative works of the software or any portion of the software, and you may copy
and distribute such modifications or works. Modified works should carry a notice stating
that you changed the software and should note the date and nature of any such change.
Please explicitly acknowledge the US Naval Research Laboratory as the original source.
This software can be redistributed and/or modified freely provided that any derivative
works bear some notice that they are derived from it, and any modified versions bear
some notice that they have been modified.

Author: David Rowenhorst;
The US Naval Research Laboratory Date: 21 Aug 2020"""

from os import environ
from pathlib import PurePath
import platform
import tempfile
from timeit import default_timer as timer

import matplotlib.pyplot as plt
import numpy as np
import pyopencl as cl

from pyebsdindex import band_detect
from pyebsdindex.opencl import openclparam
import scipy

#from os import environ
#environ['PYOPENCL_COMPILER_OUTPUT'] = '1'


RADEG = 180.0/np.pi



class BandDetect(band_detect.BandDetect):
  def __init__( self, **kwargs):
    band_detect.BandDetect.__init__(self, **kwargs)
    self.useCPU = False


  def find_bands(self, patternsIn, verbose=0, clparams=None, chunksize=528, useCPU=None, **kwargs):
    if useCPU is None:
      useCPU = self.useCPU

    if useCPU == True:
      return band_detect.BandDetect.find_bands(self, patternsIn, verbose=verbose, chunksize=-1, **kwargs)
    #if clparams is None:
    #  print('noclparams')
    #else:
    #  print(type(clparams.queue))

    try:
      tic0 = timer()
      tic = timer()
      ndim = patternsIn.ndim
      if ndim == 2:
        patterns = (np.expand_dims(patternsIn, axis=0)).copy()
      else:
        patterns = (patternsIn).copy()

      pscale = np.array([0.0, 1.0])

      if patterns.dtype.kind =='f':
        mxp = patterns.max()
        mnp = patterns.min()
        patterns -= mnp
        patterns *= (2**16-2.0)/(mxp - mnp)
        pscale[:] = np.array([mnp,(mxp - mnp) ])
        patterns = patterns.astype(np.uint16)


      shape = patterns.shape
      nPats = shape[0]

      bandData = np.zeros((nPats,self.nBands),dtype=self.dataType)
      if chunksize < 0:
        nchunks = 1
        chunksize = nPats
      else:
        nchunks = (np.ceil(nPats / chunksize)).astype(np.compat.long)

      chunk_start_end = [[i * chunksize,(i + 1) * chunksize] for i in range(nchunks)]
      chunk_start_end[-1][1] = nPats
      # these are timers used to gauge performance
      rdntime = 0.0
      convtime = 0.0
      lmaxtime = 0.0
      blabeltime = 0.0

      for chnk in chunk_start_end:
        tic1 = timer()
        nPatsChunk = chnk[1] - chnk[0]
        #rdnNorm, clparams, rdnNorm_gpu = self.calc_rdn(patterns[chnk[0]:chnk[1],:,:], clparams, use_gpu=self.CLOps[0])
        rdnNorm, clparams = self.radon_fasterCL(patterns[chnk[0]:chnk[1],:,:], padding=self.padding,
                                                                       fixArtifacts=False, background=self.backgroundsub,
                                                                       returnBuff=True, clparams=clparams)

        #rdnNorm, clparams = self.rdn_mask(rdnNorm, clparams=clparams, returnBuff=False)

        #if (self.EDAXIQ == True): # I think EDAX actually uses the convolved radon for IQ
          #nTp = self.nTheta + 2 * self.padding[1]
          #nRp = self.nRho + 2 * self.padding[0]
          #nImCL = int(rdnNorm_gpu.size/(nTp*nRp*4))
          #rdnNorm_nocov = np.zeros((nRp,nTp,nImCL),dtype=np.float32)
          #cl.enqueue_copy(clparams.queue,rdnNorm_nocov,rdnNorm,is_blocking=True)

        rdntime += timer() - tic1
        tic1 = timer()
        rdnConv, clparams = self.rdn_convCL2(rdnNorm, clparams=clparams, returnBuff=True, separableKernel=True)
        rdnNorm.release()

        convtime += timer()-tic1
        tic1 = timer()
        lMaxRdn, clparams =  self.rdn_local_maxCL(rdnConv, clparams=clparams, returnBuff=True)
        lmaxtime +=  timer()-tic1
        tic1 = timer()

        bandDataChunk = self.band_labelCL(rdnConv, lMaxRdn, clparams=clparams)
        lMaxRdn.release()
        bandData['max'][chnk[0]:chnk[1]] = bandDataChunk[0][0:nPatsChunk, :]
        bandData['avemax'][chnk[0]:chnk[1]] = bandDataChunk[1][0:nPatsChunk, :]
        bandData['maxloc'][chnk[0]:chnk[1]] = bandDataChunk[2][0:nPatsChunk, :, :]
        bandData['aveloc'][chnk[0]:chnk[1]] = bandDataChunk[3][0:nPatsChunk, :, :]
        bandData['valid'][chnk[0]:chnk[1]] = bandDataChunk[4][0:nPatsChunk, :]
        bandData['width'][chnk[0]:chnk[1]] = bandDataChunk[5][0:nPatsChunk, :]
        bandData['maxloc'][chnk[0]:chnk[1]] -= self.padding.reshape(1, 1, 2) 
        bandData['aveloc'][chnk[0]:chnk[1]] -= self.padding.reshape(1, 1, 2)
        #bandDataChunk, rdnConvBuf = self.band_label(chnk[1]-chnk[0], rdnConv, rdnNorm, lMaxRdn,
        #                                    rdnConv_gpu,rdnConv_gpu,lMaxRdn_gpu,
        #                                    use_gpu = self.CLOps[3], clparams=clparams )

        if (verbose > 1) and (chnk[1] == nPats): # need to pull the radonconv off the gpu
          nTp = self.nTheta + 2 * self.padding[1]
          nRp = self.nRho + 2 * self.padding[0]
          nImCL = int(rdnConv.size / (nTp * nRp * 4))
          rdnConvarray = np.zeros((nRp,nTp,nImCL),dtype=np.float32)

          cl.enqueue_copy(clparams.queue,rdnConvarray,rdnConv,is_blocking=True)

          rdnConvarray = rdnConvarray[:,:,0:chnk[1]-chnk[0] ]
          #plt.imshow(rdnConvarray.squeeze())

        rdnConv.release()
        rdnConv = None

        blabeltime += timer() - tic1

      bandData['avemax'] *= pscale[1]
      bandData['avemax'] += pscale[0]
      bandData['max'] *= pscale[1]
      bandData['max'] += pscale[0]
      tottime = timer() - tic0
      # going to manually clear the clparams queue -- this should clear the memory of the queue off the GPU

      #if clparams is not None:
      #  clparams.queue.finish()
      #  clparams.queue = None

      if verbose > 0:
        print('Radon Time:',rdntime)
        print('Convolution Time:', convtime)
        print('Peak ID Time:', lmaxtime)
        print('Band Label Time:', blabeltime)
        print('Total Band Find Time:',tottime)
      if verbose > 1:
        self._display_radon_pattern(rdnConvarray, bandData, patterns)
        # if len(rdnConvarray.shape) == 3:
        #   im2show = rdnConvarray[self.padding[0]:-self.padding[0],self.padding[1]:-self.padding[1], -1]
        # else:
        #   im2show = rdnConvarray[self.padding[0]:-self.padding[0],self.padding[1]:-self.padding[1]]
        #
        # rhoMaskTrim = np.int32(im2show.shape[0] * self.rhoMaskFrac)
        # mean = np.mean(im2show[rhoMaskTrim:-rhoMaskTrim, 1:-2])
        # stdv = np.std(im2show[rhoMaskTrim:-rhoMaskTrim, 1:-2])
        # im2show -= mean
        # im2show /= stdv
        # im2show = im2show.clip(-4, None)
        # im2show += 6
        # im2show[0:rhoMaskTrim,:] = 0
        # im2show[-rhoMaskTrim:,:] = 0
        #
        # im2show = np.fliplr(im2show)
        # fig = plt.figure(figsize=(12, 4))
        # subrdn = fig.add_subplot(121, xlim=(0, 180), ylim=(-self.rhoMax, self.rhoMax))
        # subrdn.imshow(
        #     im2show,
        #     cmap='gray',
        #     extent=[0, 180, -self.rhoMax, self.rhoMax],
        #     interpolation='none',
        #     zorder=1,
        #     aspect='auto'
        # )
        # width = bandData['width'][-1, :]
        # width /= width.min()
        # width *= 2.0
        # xplt = np.squeeze(180.0 - np.interp(bandData['aveloc'][-1,:,1]+0.5, np.arange(self.radonPlan.nTheta), self.radonPlan.theta))
        # yplt = np.squeeze( -1.0 * np.interp(bandData['aveloc'][-1,:,0]-0.5, np.arange(self.radonPlan.nRho), self.radonPlan.rho))
        #
        # subrdn.scatter(y=yplt, x=xplt, c='r', s=width, zorder=2)
        #
        # for pt in range(self.nBands):
        #   subrdn.annotate(str(pt + 1), np.squeeze([xplt[pt] + 4, yplt[pt]]), color='yellow')
        # #subrdn.xlim(0,180)
        # #subrdn.ylim(-self.rhoMax, self.rhoMax)
        # subpat = fig.add_subplot(122)
        # subpat.imshow(patterns[-1, :, :], cmap='gray')

    except Exception as e: # something went wrong - try the CPU
      print(e)
      bandData = band_detect.BandDetect.find_bands(self, patternsIn, verbose=verbose, chunksize=-1, **kwargs)

    return bandData


  def radon_fasterCL(self,image,padding = np.array([0,0]), fixArtifacts = False, background = None, returnBuff = True, clparams=None ):
    # this function executes the radon sumations on the GPU
    tic = timer()
    image = np.asarray(image)

    # make sure we have an OpenCL environment
    if clparams is not None:
      if clparams.queue is None:
        clparams.get_queue()
      gpu = clparams.gpu
      gpu_id = clparams.gpu_id
      ctx = clparams.ctx
      prg = clparams.prg
      queue = clparams.queue
      mf = clparams.memflags
    else:
      clparams = openclparam.OpenClParam()
      clparams.get_queue()
      gpu = clparams.gpu
      gpu_id = clparams.gpu_id
      ctx = clparams.ctx
      prg = clparams.prg
      queue = clparams.queue
      mf = clparams.memflags

    shapeIm = np.shape(image)
    if image.ndim == 2:
      nIm = 1
      image = image.reshape(1, shapeIm[0], shapeIm[1])
      shapeIm = np.shape(image)
    else:
      nIm = shapeIm[0]
    #  reform = False

    clvtypesize = 16 # this is the vector size to be used in the openCL implementation.
    nImCL = np.uint64(clvtypesize * (np.int64(np.ceil(nIm/clvtypesize))))
    imtype = image.dtype

    tict = timer()
    #image_align = np.ones((shapeIm[1], shapeIm[2], nImCL), dtype = imtype)
    #image_align[:,:,0:nIm] = np.transpose(image, [1,2,0]).copy()
    toct = timer()



    #radon_gpu = cl.Buffer(ctx,mf.READ_WRITE,size=radon.nbytes)
    #radon_gpu = cl.Buffer(ctx,mf.READ_WRITE | mf.COPY_HOST_PTR,hostbuf=radon)
    image_gpu = cl.Buffer(ctx,mf.READ_ONLY | mf.COPY_HOST_PTR,hostbuf=image)
    imstep = np.uint64(np.product(shapeIm[-2:]))
    tic = timer()

    nImChunk = np.uint64(nImCL/clvtypesize)
    image_gpuflt = cl.Buffer(ctx, mf.READ_WRITE, size=int(int(shapeIm[1])*int(shapeIm[2])*int(nImCL) * int(4)))  # 32-bit float


    if image.dtype.type is np.float32:
      prg.loadfloat32(queue, (shapeIm[2], shapeIm[1], nIm), None, image_gpu, image_gpuflt, nImCL)
    if image.dtype.type is np.ubyte:
      prg.loadubyte8(queue, (shapeIm[2], shapeIm[1], nIm), None, image_gpu, image_gpuflt, nImCL)
    if image.dtype.type is np.uint16:
      prg.loaduint16(queue, (shapeIm[2], shapeIm[1], nIm), None, image_gpu, image_gpuflt, nImCL)
    queue.flush()
    image_gpu.release()
    image_gpu = None



    if background is not None:
      back_gpu = cl.Buffer(ctx,mf.READ_ONLY | mf.COPY_HOST_PTR,hostbuf=background.astype(np.float32))
      prg.backSub(queue,(imstep, 1, 1),None,image_gpuflt,back_gpu,nImChunk)
      #imBack = np.zeros((shapeIm[1], shapeIm[2], nImCL),dtype=np.float32)
      #cl.enqueue_copy(queue,imBack,image_gpu,is_blocking=True)

    indxstep = np.uint64(self.radonPlan.indexPlan.shape[-1])
    rdnstep = np.uint64(self.nRho * self.nTheta)
    padRho = np.uint64(padding[0])
    padTheta = np.uint64(padding[1])
    shpRdn = np.asarray(((self.nRho + 2 * padding[0]), (self.nTheta + 2 * padding[1]), nImCL), dtype=np.uint64)
    radon_gpu = cl.Buffer(ctx, mf.READ_WRITE,
                          size=int((self.nRho + 2 * padding[0]) * (self.nTheta + 2 * padding[1]) * nImCL * 4))

    rdnIndx_gpu = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.radonPlan.indexPlan)
    cl.enqueue_fill_buffer(queue, radon_gpu, np.float32(self.radonPlan.missingval), 0, radon_gpu.size)
    prg.radonSum(queue,(nImChunk,rdnstep),None,rdnIndx_gpu,image_gpuflt,radon_gpu,
                  imstep, indxstep,
                 shpRdn[0], shpRdn[1],
                 padRho, padTheta, np.uint64(self.nTheta))


    if (fixArtifacts == True):
       prg.radonFixArt(queue,(nImChunk,shpRdn[0]),None,radon_gpu,
                       shpRdn[0],shpRdn[1],padTheta)

    rdnIndx_gpu.release()
    rdnIndx_gpu = None


    if returnBuff == False:
      radon = np.zeros([self.nRho + 2 * padding[0],self.nTheta + 2 * padding[1],nImCL],dtype=np.float32)
      cl.enqueue_copy(queue,radon,radon_gpu,is_blocking=True)
      radon_gpu.release()
      radon = radon[:,:, 0:nIm]
      radon_gpu = None
      #clparams = None
      return radon, clparams

    return radon_gpu, clparams



  def rdn_convCL2(self, radonIn, clparams=None, separableKernel=True, returnBuff = True):
    # this will run (eventually sequential) convolutions and identify the peaks in the convolution
    tic = timer()
    #def preparray(array):
    #  return np.require(array,None,"C")

    clvtypesize = 16  # this is the vector size to be used in the openCL implementation.

    if clparams is not None:
      if clparams.queue is None:
        clparams.get_queue()
      gpu = clparams.gpu
      gpu_id = clparams.gpu_id
      ctx = clparams.ctx
      prg = clparams.prg
      queue = clparams.queue
      mf = clparams.memflags
    else:
      clparams = openclparam.OpenClParam()
      clparams.get_queue()
      gpu = clparams.gpu
      gpu_id = clparams.gpu_id
      ctx = clparams.ctx
      prg = clparams.prg
      queue = clparams.queue
      mf = clparams.memflags

    nT = self.nTheta
    nTp = nT + 2 * self.padding[1]
    nR = self.nRho
    nRp = nR + 2 * self.padding[0]

    if isinstance(radonIn, cl.Buffer):
      rdn_gpu = radonIn
      nImCL = np.int64(rdn_gpu.size/(nTp * nRp * 4))
      shp = (nRp, nTp, nImCL)
    else:
      shp = radonIn.shape
      if len(shp) == 2:
        radon = radonIn.reshape(shp[0],shp[1],1)
        nIm = 1
      else:
        radon = radonIn
      shp = radon.shape
      nIm = shp[2]
      nImCL = np.int32(clvtypesize * (np.int64(np.ceil(nIm / clvtypesize))))
      # there is something very strange that happens if the number of images
      # is an exact multiple of the max group size (typically 256)
      mxGroupSz = gpu[gpu_id].get_info(cl.device_info.MAX_WORK_GROUP_SIZE)
      nImCL += np.int64(16 * (1 - np.int64(np.mod(nImCL,mxGroupSz) > 0)))
      radonCL = np.zeros( (nRp , nTp, nImCL), dtype = np.float32)
      radonCL[:,:,0:shp[2]] = radon
      rdn_gpu = cl.Buffer(ctx,mf.READ_ONLY | mf.COPY_HOST_PTR,hostbuf=radonCL)
      shp = (nRp, nTp, nImCL)

    nImChunk = np.uint64(nImCL / clvtypesize)
    resultConv = np.full(shp,0.0,dtype=np.float32)

    rdnConv_gpu = cl.Buffer(ctx,mf.WRITE_ONLY ,size=resultConv.nbytes)

    # maskrnd = np.zeros((self.nRho + 2 * self.padding[0], self.nTheta + 2 * self.padding[1]), dtype=np.ubyte)
    # maskrnd[self.padding[0]:-self.padding[0], self.padding[1]:-self.padding[1]] = self.rhomask1
    #
    # maskrnd = maskrnd.astype(np.ubyte)
    # maskrnd_gpu = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=maskrnd)
    #
    # prg.maskrdn(queue, (np.uint32(nT), np.uint32(nR)), None, rdn_gpu, maskrnd_gpu,
    #             np.uint64(shp[1]), np.uint64(nImChunk),
    #             np.uint64(self.padding[1]), np.uint64(self.padding[0]))

    # # pad out the radon buffers
    # prg.radonPadTheta(queue,(shp[2],shp[0],1),None,rdn_gpu,
    #                 np.uint64(shp[0]),np.uint64(shp[1]),np.uint64(self.padding[1]))

    #prg.radonPadRho2(queue,(shp[2],shp[1],1),None,rdn_gpu,
    #                  np.uint64(shp[0]),np.uint64(shp[1]),np.uint64(self.padding[0]+1))

    prg.radonPadRho2(queue, (shp[2], shp[1], 1), None, rdn_gpu,
                 np.uint64(shp[0]),np.uint64(shp[1]),np.uint64(shp[0]//2-1))

    kern_gpu = None
    if separableKernel == False:
      # for now I will assume that the kernel(s) can fit in local memory on the GPU
      # also going to assume that there is only one kernel -- this will be something to fix at some point.
      k0 = np.array(self.kernel[0,:,:], dtype=np.float32)
      kshp = np.asarray(k0.shape, dtype=np.int32)
      pad = kshp/2
      kern_gpu = cl.Buffer(ctx,mf.READ_ONLY | mf.COPY_HOST_PTR,hostbuf=k0)
      prg.convolution3d2d(queue,(np.int32((shp[1]-2*pad[1])),np.int32((shp[0]-2*pad[0])), nImChunk),None,
                        rdn_gpu, kern_gpu,np.int32(shp[1]),np.int32(shp[0]),np.int32(shp[2]),
                        np.int32(kshp[1]), np.int32(kshp[0]), np.int32(pad[1]), np.int32(pad[0]), rdnConv_gpu)



      #tic = timer()
    else: # convolution is separable
      tempConvbuff = cl.Buffer(ctx,mf.HOST_NO_ACCESS,size=(shp[0]*shp[1]*shp[2]*4))

      kshp = np.asarray(self.kernel[0,:,:].shape,dtype=np.int32)
      pad = kshp
      k0x = np.require(self.kernel[0, np.int64(kshp[0] / 2), :], requirements=['C', 'A', 'W', 'O'], dtype=np.float32)
      k0x *= 1.0 / k0x.sum()
      k0x = (k0x[...,:]).reshape(1,kshp[1])



      kshp = np.asarray(k0x.shape,dtype=np.int32)

      kern_gpu_x = cl.Buffer(ctx,mf.READ_ONLY | mf.COPY_HOST_PTR,hostbuf=k0x)
      prg.convolution3d2d(queue,(np.int32((shp[1]-2*pad[1])),np.int32((shp[0]-2*pad[0])), nImChunk),None,
                          rdn_gpu,kern_gpu_x,np.int32(shp[1]),np.int32(shp[0]),np.int32(shp[2]),
                          np.int32(kshp[1]),np.int32(kshp[0]),np.int32(pad[1]),np.int32(pad[0]),tempConvbuff)

      kshp = np.asarray(self.kernel[0,:,:].shape,dtype=np.int32)
      k0y = np.require(self.kernel[0, :, np.int32(kshp[1] / 2)], requirements=['C', 'A', 'W', 'O'], dtype=np.float32)
      k0y *= 1.0 / k0y.sum()
      k0y = (k0y[...,:]).reshape(kshp[0],1)
      kshp = np.asarray(k0y.shape,dtype=np.int32)

      kern_gpu_y = cl.Buffer(ctx,mf.READ_ONLY | mf.COPY_HOST_PTR,hostbuf=k0y)
      prg.convolution3d2d(queue,(np.int32((shp[1]-2*pad[1])),np.int32((shp[0]-2*pad[0])), nImChunk),None,
                          tempConvbuff,kern_gpu_y,np.int32(shp[1]),np.int32(shp[0]),np.int32(shp[0]),
                          np.int32(kshp[1]),np.int32(kshp[0]),np.int32(pad[1]),np.int32(pad[0]),rdnConv_gpu)

    # for each radon, get the min value
    mns = cl.Buffer(ctx,mf.READ_WRITE,size=nImCL * 4)

    prg.imageMin(queue,(nImChunk,1,1),None,
                 rdnConv_gpu, mns,np.uint32(shp[1]),np.uint32(shp[0]),
                 np.uint32(self.padding[1]),np.uint32(self.padding[0]))
    # subtract the min value, clipping to 0.
    prg.imageSubMinWClip(queue,(np.int32(shp[1]), np.int32(shp[0]),nImChunk),None,
                     rdnConv_gpu,mns,np.uint32(shp[1]),np.uint32(shp[0]),
                     np.uint32(0),np.uint32(0))



    #rdn_gpu.release()
    mns.release()
    if kern_gpu is None:
      kern_gpu_y.release()
      kern_gpu_x.release()
      tempConvbuff.release()
    else:
      kern_gpu.release()

    #cl.enqueue_copy(queue,resultConv,rdnConv_gpu,is_blocking=True)

    if returnBuff == False:
      cl.enqueue_copy(queue, resultConv, rdnConv_gpu, is_blocking=True)
      rdnConv_gpu.release()
      rdnConv_gpu = None
      return resultConv, clparams
    else:
      return rdnConv_gpu, clparams



  def rdn_local_maxCL(self,radonIn, clparams=None,  returnBuff = True):
    # this will run a morphological max kernel over the convolved radon array
    # the local max identifies the location of the peak max within
    # the window size.

    tic = timer()
    clvtypesize = 16  # this is the vector size to be used in the openCL implementation.


    mf = cl.mem_flags
    if clparams is not None:
      if clparams.queue is None:
        clparams.get_queue()
      gpu = clparams.gpu
      gpu_id = clparams.gpu_id
      ctx = clparams.ctx
      prg = clparams.prg
      queue = clparams.queue
      mf = clparams.memflags
    else:
      clparams = openclparam.OpenClParam()
      clparams.get_queue()
      gpu = clparams.gpu
      gpu_id = clparams.gpu_id
      ctx = clparams.ctx
      prg = clparams.prg
      queue = clparams.queue
      mf = clparams.memflags



    nT = self.nTheta
    nTp = nT + 2 * self.padding[1]
    nR = self.nRho
    nRp = nR + 2 * self.padding[0]

    if isinstance(radonIn,cl.Buffer):
      rdn_gpu = radonIn
      nImCL = np.int64(rdn_gpu.size / (nTp * nRp * 4))
      shp = (nRp,nTp,nImCL)
    else:
      shp = radonIn.shape
      if len(shp) == 2:
        radon = radonIn.reshape(shp[0],shp[1],1)
      else:
        radon = radonIn
      shp = radon.shape
      nIm = shp[2]
      nImCL = np.int32(clvtypesize * (np.int64(np.ceil(nIm / clvtypesize))))
      # there is something very strange that happens if the number of images
      # is a exact multiple of the max group size (typically 256)
      mxGroupSz = gpu[gpu_id].get_info(cl.device_info.MAX_WORK_GROUP_SIZE)
      nImCL += np.int64(16 * (1 - np.int64(np.mod(nImCL,mxGroupSz) > 0)))
      radonCL = np.zeros((nRp,nTp,nImCL),dtype=np.float32)
      radonCL[:,:,0:shp[2]] = radon
      rdn_gpu = cl.Buffer(ctx,mf.READ_ONLY | mf.COPY_HOST_PTR,hostbuf=radonCL)
      shp = (nRp,nTp,nImCL)


    nImChunk = np.uint64(nImCL / clvtypesize)
    #out = np.zeros((shp), dtype = np.int32)

    lmaxX = cl.Buffer(ctx, mf.READ_WRITE, size=rdn_gpu.size)
    lmaxXY = cl.Buffer(ctx, mf.READ_WRITE, size=rdn_gpu.size)


    nChunkT = np.uint64(np.ceil(nT / np.float64(self.peakPad[1])))
    nChunkR = np.uint64(np.ceil(nR / np.float64(self.peakPad[0])))
    winszX = np.uint64(self.peakPad[1])
    winszX2 = np.uint64((winszX-1) / 2)
    winszY = np.uint64(self.peakPad[0])
    winszY2 = np.uint64((winszY-1) / 2)


    # prg.morphDilateKernelBF(queue,(np.uint32(nT),np.uint32(nR),nImChunk),None,rdn_gpu,lmaxX,
    #                         np.int64(shp[1]),np.int64(shp[0]),
    #                         np.int64(self.padding[1]),np.int64(self.padding[0]),
    #                         np.int64(1),np.int64(self.peakPad[0]))
    #
    # prg.morphDilateKernelBF(queue,(np.uint32(nT),np.uint32(nR),nImChunk),None,lmaxX,lmaxXY,
    #                         np.int64(shp[1]),np.int64(shp[0]),
    #                         np.int64(self.padding[1]),np.int64(self.padding[0]),
    #                         np.int64(self.peakPad[1]),np.int64(1))
    # calculate the max in the x direction
    prg.morphDilateKernelBF(queue, (np.uint32(shp[1]), np.uint32(nR), nImChunk), None, rdn_gpu, lmaxX,
                            np.int64(shp[1]), np.int64(shp[0]),
                            np.int64(0), np.int64(self.padding[0]),
                            np.int64(self.peakPad[1]), np.int64(1))
    # take the max in the x output, use as input, and calculate in the y direction
    prg.morphDilateKernelBF(queue, (np.uint32(nT), np.uint32(nR), nImChunk), None, lmaxX, lmaxXY,
                            np.int64(shp[1]), np.int64(shp[0]),
                            np.int64(self.padding[1]), np.int64(self.padding[0]),
                            np.int64(1), np.int64(self.peakPad[0]))

    local_max = np.zeros((shp),dtype=np.ubyte)
    local_max_gpu = cl.Buffer(ctx,mf.WRITE_ONLY,size=local_max.nbytes)

    prg.im1EQim2(queue,(np.uint32(nT),np.uint32(nR),nImCL),None, lmaxXY, rdn_gpu, local_max_gpu,
                 np.uint64(shp[1]),np.uint64(shp[0]),
                 np.uint64(self.padding[1]),np.uint64(self.padding[0]))

    maskrnd = np.zeros((self.nRho + 2 * self.padding[0] , self.nTheta + 2 * self.padding[1]), dtype=np.ubyte)
    maskrnd[self.padding[0]:-self.padding[0], self.padding[1]:-self.padding[1]] = self.rdnmask.astype(np.ubyte)

    maskrnd = maskrnd.astype(np.ubyte)
    maskrnd_gpu = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=maskrnd)

    prg.maxmask(queue, (np.uint32(nT), np.uint32(nR)), None, local_max_gpu, maskrnd_gpu,
                 np.uint64(shp[1]), np.uint64(nImChunk),
                 np.uint64(self.padding[1]), np.uint64(self.padding[0]))

    queue.flush()


    if returnBuff == False:
      local_maxX = np.zeros((shp), dtype=np.float32)
      local_maxXY = np.zeros((shp), dtype=np.float32)
      cl.enqueue_copy(queue,local_max,local_max_gpu,is_blocking=True)
      cl.enqueue_copy(queue, local_maxX, lmaxX, is_blocking=True)
      cl.enqueue_copy(queue, local_maxXY, lmaxXY, is_blocking=True)
      queue.flush()
      rdn_gpu.release()
      lmaxX.release()
      lmaxXY.release()
      local_max_gpu.release()
      local_max_gpu = None
      rhoMaskTrim = np.int32((shp[0] - 2 * self.padding[0]) * self.rhoMaskFrac + self.padding[0])
      local_max[0:rhoMaskTrim,:, :] = 0
      local_max[-rhoMaskTrim:,:,:] = 0
      local_max[:,0:self.padding[1],:] = 0
      local_max[:,-self.padding[1]:,:] = 0
      return (local_max,local_maxX, local_maxXY) , clparams
    else:
      lmaxX.release()
      lmaxXY.release()
      return local_max_gpu, clparams


  def band_labelCL(self,rdnConvIn, lMaxRdnIn,clparams=None):

    # an attempt to run the band label on the GPU

    tic = timer()


    if clparams is not None:
      if clparams.queue is None:
        clparams.get_queue()
      gpu = clparams.gpu
      ctx = clparams.ctx
      prg = clparams.prg
      queue = clparams.queue
      mf = clparams.memflags
    else:
      clparams = openclparam.OpenClParam()
      clparams.get_queue()
      gpu = clparams.gpu
      ctx = clparams.ctx
      prg = clparams.prg
      queue = clparams.queue
      mf = clparams.memflags


    nT = self.nTheta
    nTp = nT + 2 * self.padding[1]
    nR = self.nRho
    nRp = nR + 2 * self.padding[0]

    if isinstance(rdnConvIn,cl.Buffer):
      rdnConv_gpu = rdnConvIn
      nIm = np.int64(rdnConv_gpu.size / (nTp * nRp * 4))
      shp = (nRp,nTp,nIm)
    else:
      shp = rdnConvIn.shape
      if len(shp) == 2:
        radonConv = rdnConvIn.reshape(shp[0],shp[1],1)
      else:
        radonConv = rdnConvIn
      shp = radonConv.shape
      nIm = shp[2]

      radonConvCL = np.zeros((nRp,nTp,nIm),dtype=np.float32)
      radonConvCL[:,:,0:shp[2]] = radonConv
      rdnConv_gpu = cl.Buffer(ctx,mf.READ_ONLY | mf.COPY_HOST_PTR,hostbuf=radonConvCL)
      shp = (nRp,nTp,nIm)

    if isinstance(lMaxRdnIn,cl.Buffer):
      lMaxRdn_gpu = lMaxRdnIn
      nIm = np.int64(lMaxRdn_gpu.size / (nTp * nRp))
      shp = (nRp,nTp,nIm)
    else:
      shp = lMaxRdnIn.shape
      if len(shp) == 2:
        lMaxRdn = lMaxRdnIn.reshape(shp[0],shp[1],1)
      else:
        lMaxRdn = lMaxRdnIn
      shp = lMaxRdnIn.shape
      nIm = shp[2]

      lMaxRdnCL = np.zeros((nRp,nTp,nIm),dtype=np.float32)
      lMaxRdnCL[:,:,0:shp[2]] = lMaxRdn
      lMaxRdn_gpu = cl.Buffer(ctx,mf.READ_ONLY | mf.COPY_HOST_PTR,hostbuf=lMaxRdnCL)
      shp = (nRp,nTp,nIm)


    maxval = np.zeros((nIm, self.nBands),dtype=np.float32)-2.0e6
    maxval_gpu = cl.Buffer(ctx,mf.WRITE_ONLY,size=maxval.nbytes)
    maxloc = np.zeros((nIm, self.nBands),dtype=np.int64)
    maxloc_gpu = cl.Buffer(ctx,mf.WRITE_ONLY,size=maxloc.nbytes)
    aveval = np.zeros((nIm,self.nBands),dtype=np.float32) - 2.0e6
    aveval_gpu = cl.Buffer(ctx,mf.WRITE_ONLY,size=aveval.nbytes)
    width = np.zeros((nIm, self.nBands), dtype=np.float32)
    width_gpu = cl.Buffer(ctx, mf.WRITE_ONLY, size=width.nbytes)
    aveloc = np.zeros((nIm,self.nBands,2),dtype=np.float32)
    aveloc_gpu = cl.Buffer(ctx,mf.WRITE_ONLY,size=aveloc.nbytes)
    #rhoMaskTrim = np.int64((shp[0] - 2 * self.padding[0]) * self.rhoMaskFrac + self.padding[0])
    rhoMaskTrim = np.int64(self.padding[0])

    prg.maxlabel(queue,(nIm, 1,1),(1,1,1),
                 lMaxRdn_gpu,rdnConv_gpu,
                 maxloc_gpu, maxval_gpu,
                 aveloc_gpu,aveval_gpu,
                 width_gpu,
                 np.int64(shp[1]),np.int64(shp[0]),
                 np.int64(self.padding[1]),rhoMaskTrim,np.int64(self.nBands) )


    queue.finish()
    cl.enqueue_copy(queue,maxval,maxval_gpu,is_blocking=False)
    cl.enqueue_copy(queue,maxloc,maxloc_gpu,is_blocking=False)
    cl.enqueue_copy(queue,aveval,aveval_gpu,is_blocking=False)
    cl.enqueue_copy(queue,aveloc,aveloc_gpu,is_blocking=False)
    cl.enqueue_copy(queue, width, width_gpu, is_blocking=True)

    #rdnConv_out = np.zeros((nRp,nTp,nIm),dtype=np.float32)
    #cl.enqueue_copy(queue,rdnConv_out,rdnConv_gpu,is_blocking=True)
    #queue.finish()
    #rdnConv_gpu.release()
    maxval_gpu.release()
    maxloc_gpu.release()
    aveval_gpu.release()
    aveloc_gpu.release()
    width_gpu.release()
    queue.finish()

    maxlocxy = np.zeros((nIm, self.nBands, 2),dtype=np.float32)
    temp = np.asarray(np.unravel_index(maxloc, (shp[0], shp[1])), dtype = np.float32)
    maxlocxy[:,:,0] = temp[0,:,:]
    maxlocxy[:,:,1] = temp[1,:,:]
    valid = np.asarray(maxval > -1e6, dtype=np.int8)



    return (maxval,aveval,maxlocxy,aveloc,valid, width)

def getopenclparam(**kwargs):
  clparam = openclparam.OpenClParam(**kwargs)

  return clparam



# This software was developed by employees of the US Naval Research Laboratory (NRL), an
# agency of the Federal Government. Pursuant to title 17 section 105 of the United States
# Code, works of NRL employees are not subject to copyright protection, and this software
# is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
# responsibility whatsoever for its use by other parties, and makes no guarantees,
# expressed or implied, about its quality, reliability, or any other characteristic. We
# would appreciate acknowledgment if the software is used. To the extent that NRL may hold
# copyright in countries other than the United States, you are hereby granted the
# non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
# works and distribute this software, in any medium, or authorize others to do so on your
# behalf, on a royalty-free basis throughout the world. You may improve, modify, and
# create derivative works of the software or any portion of the software, and you may copy
# and distribute such modifications or works. Modified works should carry a notice stating
# that you changed the software and should note the date and nature of any such change.
# Please explicitly acknowledge the US Naval Research Laboratory as the original source.
# This software can be redistributed and/or modified freely provided that any derivative
# works bear some notice that they are derived from it, and any modified versions bear
# some notice that they have been modified.
#
# Author: David Rowenhorst;
# The US Naval Research Laboratory Date: 21 Aug 2020

import numpy as np

from pyebsdindex import ebsd_index, pcopt


class TestPCOptimization:
    def test_pc_optimize(self, pattern_al_sim_20kv):
        pc0 = (0.4, 0.72, 0.6)
        indexer = ebsd_index.EBSDIndexer(patDim=pattern_al_sim_20kv.shape)
        new_pc = pcopt.optimize(pattern_al_sim_20kv, indexer, PC0=pc0)
        assert np.allclose(new_pc, pc0, atol=0.05)

    def test_pc_optimize_pso(self, pattern_al_sim_20kv):
        pc0 = (0.4, 0.72, 0.6)
        indexer = ebsd_index.EBSDIndexer(patDim=pattern_al_sim_20kv.shape)
        new_pc = pcopt.optimize_pso(pattern_al_sim_20kv, indexer, PC0=pc0)
        assert np.allclose(new_pc, pc0, atol=0.05)



'''This software was developed by employees of the US Naval Research Laboratory (NRL), an
agency of the Federal Government. Pursuant to title 17 section 105 of the United States
Code, works of NRL employees are not subject to copyright protection, and this software
is in the public domain. PyEBSDIndex is an experimental system. NRL assumes no
responsibility whatsoever for its use by other parties, and makes no guarantees,
expressed or implied, about its quality, reliability, or any other characteristic. We
would appreciate acknowledgment if the software is used. To the extent that NRL may hold
copyright in countries other than the United States, you are hereby granted the
non-exclusive irrevocable and unconditional right to print, publish, prepare derivative
works and distribute this software, in any medium, or authorize others to do so on your
behalf, on a royalty-free basis throughout the world. You may improve, modify, and
create derivative works of the software or any portion of the software, and you may copy
and distribute such modifications or works. Modified works should carry a notice stating
that you changed the software and should note the date and nature of any such change.
Please explicitly acknowledge the US Naval Research Laboratory as the original source.
This software can be redistributed and/or modified freely provided that any derivative
works bear some notice that they are derived from it, and any modified versions bear
some notice that they have been modified.

Author: David Rowenhorst;
The US Naval Research Laboratory Date: 21 Aug 2020'''



import matplotlib.colors as pltcolors
import matplotlib.pyplot as plt
import numpy as np

from pyebsdindex import rotlib


def makeipf(ebsddata, indexer, vector=np.array([0,0,1.0]), xsize = None, ysize = None):
  nphase = len(indexer.phaseLib)

  npoints = ebsddata.shape[-1]
  ipfphase = np.zeros((nphase,npoints,3), dtype =np.float32)+1

  phcount = 0
  for ph in indexer.phaseLib:
    quat = ebsddata[phcount]['quat']
    if ph.lauecode == 43:
      ipfphase[phcount, :, :] = qu2ipf_cubic(quat, vector=vector)
    if ph.lauecode == 62:
      ipfphase[phcount, :, :] = qu2ipf_hex(quat, vector=vector)
    phcount += 1
  phase = ((ebsddata[-1]['phase']).copy()).clip(0).reshape(npoints,1)
  ipfout = np.choose(phase, ipfphase).squeeze()
  ipfout[ebsddata[-1]['fit'] > 179,:] = 0


  if xsize is not None:
    xsize = int(xsize)
    #if ysize is None:
      #print(ysize)
  else:
    xsize = indexer.fID.nCols
    #xsize = int(npoints)
    #ysize = 1

  if ysize is not None:
    ysize = int(ysize)
  else:
    ysize = int(npoints // xsize + np.int64((npoints % xsize) > 0))


  ipf_out = np.zeros((ysize, xsize,3), dtype=np.float32)
  ipf_out = ipf_out.flatten()
  npts = min(int(npoints), int(xsize*ysize))
  # if int(xsize*ysize) < npoints:
  #   npts = int(xsize*ysize)
  ipf_out[0:npts*3] = ipfout[0:npts,:].flatten()
  ipf_out = ipf_out.reshape(ysize, xsize, 3)
  return ipf_out



def qu2ipf_cubic(quats, vector=np.array([0,0,1.0])):
  xstalvect = rotlib.quat_vector(quats,vector)
  return ipf_color_cubic(xstalvect).clip(0.0, 1.0)

def ipf_color_cubic(xstalvect):
  shp = xstalvect.shape
  if len(shp) == 1:
    xstalv = np.copy(xstalvect.reshape(1,shp))
  else:
    xstalv = np.copy(xstalvect)
  npoints = shp[0]

  xstalv = np.abs(xstalv)
  xstalv = np.sort(xstalv, axis=1)
  # project to sterographic triangle
  xP = (2 * xstalv[:,1]) / (1 + xstalv[:,2])
  yP = (2 * xstalv[:,0]) / (1 + xstalv[:,2])

  # cubic unit tri center
  triPts = np.array( [[0,0],
                      [2. / np.sqrt(2.) / (1. + 1. / np.sqrt(2.)),0],
                      [2. / np.sqrt(3.) / (1. + 1. / np.sqrt(3.)),2. / np.sqrt(3.) / (1. + 1. / np.sqrt(3.))]], dtype = np.float32)

  middle = np.tan(1. / 2. * np.arctan(triPts[2,1] / triPts[2,0]))
  a = np.sqrt( (triPts[2,1] - triPts[1,1]) ** 2. + (triPts[2,0] - triPts[1,0]) ** 2.)
  b = np.sqrt((triPts[1,0]) ** 2. + (triPts[1,1]) ** 2.)
  c = np.sqrt(triPts[2,0] ** 2. + triPts[2,1] ** 2.)

  #y0 = 1/2. * np.sqrt( ((b+c-a)*(c+a-b)*(a+b-c)) / (a+b+c) )
  #x0 = y0 / middle
  y0 = np.mean(triPts[:, 1])
  x0 = np.mean(triPts[:, 0])

  S = np.sqrt((xP - x0) ** 2. + (yP - y0) ** 2.)
  H = np.arctan2((yP - y0) , (xP - x0)) *180.0/np.pi
  V = np.ones(npoints)

  #H =  (xP < x0).astype(np.float)*180.0+H
  H = H + 240.0 - np.arctan2((triPts[2,1] - y0) , (triPts[2,0] - x0)) * 180.0/np.pi
  #H = H - np.arctan2(-y0 , -x0) * 180.0 / np.pi
  sMax = np.sqrt(x0**2+y0**2)
  S = S / (sMax) * 0.8 + 0.2


  H = H % (360.0)
  H = H / 360.0

  RGB = pltcolors.hsv_to_rgb(np.array([H,S,V]).T)

  return RGB

def ipf_ledgend_cubic_old(size=512):
  szx = size
  aspect = 0.94
  szy = np.round(size*aspect).astype(int)
  triangleWX = np.round(size*0.86).astype(int)
  triangleWY = np.round(triangleWX * aspect).astype(int)

  triOrigin = np.round(np.array([0.1,0.1])*size).astype(int)

  triScale = 0.82842708/triangleWX
  np0 = triangleWX*triangleWY
  triXY = np.indices([triangleWY,triangleWX])
  triYX_stereo = (triXY*triScale).reshape(2,np0)
  xt = triYX_stereo[1,:]
  yt = triYX_stereo[0,:]
  xyz = np.zeros((np0, 3))
  xyz[:,2] = (4. - (xt**2.+ yt**2.))/(4. + (xt**2.+ yt**2.))
  xyz[:,0] = (xyz[:,2] + 1.) * (xt / 2.)
  xyz[:,1] = (xyz[:,2] + 1.) * (yt / 2.)
  pltest = (np.sqrt(2.0) * (-xyz[:,0]+xyz[:,2])) >= 0.0

  wh = np.nonzero( (xyz[:,1] <= xyz[:,0]) & pltest )[0]
  rgbaTri = np.full((np0, 4), 1.0, dtype = np.float32)
  rgbaTri[:,3] = 0
  rgbaTri[wh,3] = 1.0
  rgbaTri[wh,0:3] = ipf_color_cubic(xyz[wh,:])
  rgbaTri = rgbaTri.reshape(triangleWY,triangleWX,4)

  palette = np.full((szy,szx, 4), 1.0)
  trim = palette.shape[0] - triOrigin[0] - rgbaTri.shape[0]
  palette[triOrigin[1]:triOrigin[1]+triangleWY, triOrigin[0]:triOrigin[0]+triangleWX,:] = rgbaTri[:, :,:]

  dpi = size
  fsize = 6.0

  fig = plt.figure(figsize=(1.0,1.0),dpi=size/1.0)
  ax = plt.Axes(fig,[0.,0.,1.,aspect])
  ax.set_axis_off()
  fig.add_axes(ax)
  img = plt.imshow(palette, origin='lower', extent=[0,szx,0,szy])
  anno001 = plt.text(triOrigin[0] - 8*fsize*size/512,triOrigin[1] - 6*fsize*size/512.0, '001', fontsize = fsize)
  anno011 = plt.text(size - 14*fsize*size/512,triOrigin[1] - 6*fsize*size/512.0,'011',fontsize=fsize)
  anno111 = plt.text(size - 14*fsize*size/512,(triangleWY+triOrigin[1])*0.95,'111',fontsize=fsize)
  #ax = img.gca()
  fig.savefig("IPFCubic.pdf",bbox_inches='tight')
  #plt.show()
  #img.axes.get_xaxis().set_visible(False)
  #img.axes.get_yaxis().set_visible(False)

  #plt.savefig("test.png",bbox_inches='tight')
  #plt.savefig("test.png",bbox_inches = 0)

def ipf_ledgend_cubic(size=512):
  szx = size
  aspect = 0.89
  szy = np.round(size*aspect).astype(int)
  triangleWX = np.round(size*1.0).astype(int)
  triangleWY = np.round(triangleWX * aspect).astype(int)

  #triOrigin = np.round(np.array([0.1,0.1])*size).astype(int)
  triOrigin = np.array([0,0]).astype(int)

  triScale = 0.82842708/triangleWX
  np0 = triangleWX*triangleWY
  triXY = np.indices([triangleWY,triangleWX])
  triYX_stereo = (triXY*triScale).reshape(2,np0)
  xt = triYX_stereo[1,:]
  yt = triYX_stereo[0,:]
  xyz = np.zeros((np0, 3))
  xyz[:,2] = (4. - (xt**2.+ yt**2.))/(4. + (xt**2.+ yt**2.))
  xyz[:,0] = (xyz[:,2] + 1.) * (xt / 2.)
  xyz[:,1] = (xyz[:,2] + 1.) * (yt / 2.)
  pltest = (np.sqrt(2.0) * (-xyz[:,0]+xyz[:,2])) >= 0.0

  wh = np.nonzero( (xyz[:,1] <= xyz[:,0]) & pltest )[0]
  rgbaTri = np.full((np0, 4), 1.0, dtype = np.float32)
  rgbaTri[:,3] = 0.0
  rgbaTri[wh,3] = 1.0
  rgbaTri[wh,0:3] = ipf_color_cubic(xyz[wh,:])
  rgbaTri = rgbaTri.reshape(triangleWY,triangleWX,4)
  dpi = size
  figsz = 1.0
  fsize = 5.0#/512*size

  fig = plt.figure(1001, figsize=(figsz,figsz),dpi=size/figsz*0.5)
  ax = plt.Axes(fig,[0.05,0.1,0.9,aspect*0.9])
  ax.set_axis_off()
  fig.add_axes(ax)

  img = plt.imshow(rgbaTri, origin='lower', extent=[0,szx,0,szy])
  anno001 = plt.text(triOrigin[0] - 5*fsize*size/512/figsz,triOrigin[1] - 7*fsize*size/512.0/figsz, '001', fontsize = fsize)
  anno011 = plt.text(size - 10*fsize*size/512/figsz,triOrigin[1] - 7*fsize*size/512.0/figsz,'011',fontsize=fsize)
  anno111 = plt.text(size - 10*fsize*size/512/figsz,(triangleWY+triOrigin[1])*1.0,'111',fontsize=fsize)
  fig.savefig("IPFCubic.pdf",bbox_inches=0, transparent=True)
  plt.close(1001)


def qu2ipf_hex(quats, vector=np.array([0,0,1.0])):
  xstalvect = rotlib.quat_vector(quats,vector)
  return ipf_color_hex(xstalvect).clip(0.0, 1.0)

def ipf_color_hex(xstalvect):
  shp = xstalvect.shape
  if len(shp) == 1:
    xstalv = np.copy(xstalvect.reshape(1,shp))
  else:
    xstalv = np.copy(xstalvect)
  npoints = shp[0]

  xstalv /= np.sqrt((xstalv ** 2).sum(-1))[..., np.newaxis]
  xstalv[xstalv[:, 2] < 0, :] *= -1

  theta = np.arctan2(xstalv[:,1], xstalv[:,0])
  wh = np.where(theta >= np.pi/3.)[0]
  q60 = rotlib.quatnorm(np.array([ np.cos(np.pi/6.0),0, 0, -0.50000000]))

  while wh.size > 0:
    xstalv[wh,:] = rotlib.quat_vector(q60,xstalv[wh,:] )
    theta = np.arctan2(xstalv[:, 1], xstalv[:, 0])
    wh = np.where(theta >= np.pi / 3.)[0]



  theta = np.arctan2(xstalv[:, 1], xstalv[:, 0])
  wh = np.where(theta < 0.0)[0]
  q60 = np.array([np.cos(np.pi / 6.0), 0, 0, 0.50000000])
  while wh.size > 0:
    xstalv[wh, :] = rotlib.quat_vector(q60, xstalv[wh, :])
    theta = np.arctan2(xstalv[:, 1], xstalv[:, 0])
    wh = np.where(theta < 0.0)[0]


  theta = np.arctan2(xstalv[:, 1], xstalv[:, 0])
  wh = np.where(theta >= np.pi / 6.)[0]
  if wh.size > 0:
    nx = -np.sin(np.pi / 6.)
    ny = np.cos(np.pi / 6.)
    const = 2. * (nx * xstalv[wh,0] + ny * xstalv[wh, 1])
    xstalv[wh, 0] -= const * nx
    xstalv[wh, 1] -= const * ny



  xP = (xstalv[:,0]) / (1 + xstalv[:,2])
  yP = (xstalv[:,1]) / (1 + xstalv[:,2])

  # cubic unit tri center
  triPts = np.array( [[0,0],
                      [1.0 ,0],
                      [np.sqrt(3.)/2.0, 0.5 ]], dtype = np.float32)

  middle = np.tan(1. / 2. * np.arctan(triPts[2,1] / triPts[2,0]))

  a = np.sqrt( (triPts[2,1] - triPts[1,1]) ** 2. + (triPts[2,0] - triPts[1,0]) ** 2.)
  b = np.sqrt((triPts[1,0]) ** 2. + (triPts[1,1]) ** 2.)
  c = np.sqrt(triPts[2,0] ** 2. + triPts[2,1] ** 2.)

  #y0 = 0.4 * np.sqrt( ((b+c-a)*(c+a-b)*(a+b-c)) / (a+b+c) )
  #x0 = y0 / middle
  y0 = np.mean(triPts[:,1])
  x0 = np.mean(triPts[:, 0])


  S = np.sqrt((xP - x0) ** 2. + (yP - y0) ** 2.)
  H = np.arctan2((yP - y0) , (xP - x0)) * 180.0 / np.pi
  V = np.ones(npoints)

  #H = (xP < x0).astype(np.float) * 180.0 + H
  H = H + 240 - np.arctan2((triPts[2, 1] - y0) , (triPts[2, 0] - x0)) * 180.0 / np.pi
  #H = H  - np.arctan2((- y0), ( - x0)) * 180.0 / np.pi
  sMax = np.sqrt((triPts[:,0] - x0) ** 2 + (triPts[:,1] - y0) ** 2).max()
  S = S / (sMax) * 0.75 + 0.25

  H = H % (360.0)
  H = H / 360.0

  RGB = pltcolors.hsv_to_rgb(np.array([H,S,V]).T)

  return RGB


def ipf_ledgend_hex(size=512):
  szx = size
  aspect = 0.6
  szy = np.round(size*aspect).astype(int)
  triangleWX = np.round(size*1.0).astype(int)
  triangleWY = np.round(triangleWX * aspect).astype(int)

  #triOrigin = np.round(np.array([0.1,0.1])*size).astype(int)
  triOrigin = np.array([0,0]).astype(int)

  triScale = 1.0/triangleWX #0.82842708/triangleWX
  np0 = triangleWX*triangleWY
  triXY = np.indices([triangleWY,triangleWX])
  triYX_stereo = (triXY*triScale).reshape(2,np0)
  xt = triYX_stereo[1,:]*2
  yt = triYX_stereo[0,:]*2

  xyz = np.zeros((np0, 3))

  xyz[:,2] = (4. - (xt**2+ yt**2))/(4. + (xt**2+ yt**2))
  xyz[:,0] = (xyz[:,2] + 1.) * (xt / 2.)
  xyz[:,1] = (xyz[:,2] + 1.) * (yt / 2.)
  pltest = np.sqrt(xt ** 2 + yt**2) < 2.0
  pltest2 = (xyz[:,2] >= 0.0).squeeze()

  theta  =  np.arctan2(xyz[:,1],xyz[:,0]) < np.pi/6
  theta = np.logical_and( theta, pltest)
  #theta = np.logical_and(theta, pltest2)

  wh = np.nonzero( theta)[0]
  #return xyz[wh,:]
  rgbaTri = np.full((np0, 4), 1.0, dtype = np.float32)
  rgbaTri[:,3] = 0.0
  rgbaTri[wh,3] = 1.0
  rgbaTri[wh,0:3] = ipf_color_hex(xyz[wh,:])

  rgbaTri = rgbaTri.reshape(triangleWY,triangleWX,4)
  dpi = size
  figsz = 1.0
  fsize = 4.0#/512*size

  fig = plt.figure(1001, figsize=(figsz,figsz*aspect),dpi=size/figsz*0.5)
  ax = plt.Axes(fig,[-0.2,0.15,1.4,aspect*1.4])
  ax.set_axis_off()
  fig.add_axes(ax)

  img = plt.imshow(rgbaTri, origin='lower', extent=[0,szx,0,szy])
  anno001 = plt.text(triOrigin[0] - 5*fsize*size/512/figsz,triOrigin[1] - 8*fsize*size/512.0/figsz, r'0001', fontsize = 0.9*fsize)
  anno011 = plt.text(size - 10*fsize*size/512/figsz,triOrigin[1] - 9.5*fsize*size/512.0/figsz,r'$2\bar{1}\bar{1}0$',fontsize=0.9*fsize)
  anno111 = plt.text(size - 25*fsize*size/512/figsz,(triangleWY+triOrigin[1])*0.85,r'$10\bar{1}0$',fontsize=0.9*fsize)
  fig.savefig("IPFHex.pdf",bbox_inches=0, transparent=True)
  plt.close(1001)